"""
vtreat main implementation
"""

import abc
import math
import pprint
import warnings
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple

import numpy
import pandas

import vtreat.util
import vtreat.transform

import sklearn.base


bad_sentinel = "_NA_"


def replace_bad_with_sentinel(ar: List) -> numpy.ndarray:
    """
    Replace None/NaN entries in iterable with '_NA_'
    :param ar: iterable
    :return: one dimensional numpy array
    """

    ar = numpy.array(ar)
    bads = pandas.isnull(ar)
    ar[bads] = bad_sentinel
    return ar


def ready_data_frame(d) -> Tuple[pandas.DataFrame, type]:
    """
    Convert an array-like object to a data frame for processing.

    :param d: data frame like object to work with
    :return: dataframe with string-named columns
    """
    orig_type = type(d)
    if not isinstance(d, pandas.DataFrame):
        d = pandas.DataFrame(d)
    else:
        d = d.copy()
    d.columns = [str(c) for c in d.columns]  # convert column names to strings
    return d, orig_type


def back_to_orig_type_data_frame(
    d: pandas.DataFrame, orig_type: type
) -> Tuple[Any, List[str]]:
    """
    Convert data frame back to ndarray if that was the original type.

    :param d: data frame
    :param orig_type: type of original object
    :return: converted result
    """
    if not isinstance(d, pandas.DataFrame):
        raise TypeError(
            "Expected result to be a pandas.DataFrame, found: " + str(type(d))
        )
    columns = [c for c in d.columns]
    if orig_type == numpy.ndarray:
        d = numpy.asarray(d)
    return d, columns


class VarTransform(abc.ABC):
    """
    Base class for vtreat transforms
    """

    incoming_column_name_: str
    incoming_column_is_numeric_: bool
    derived_column_names_: List[str]
    need_cross_treatment_: bool
    treatment_: str
    refitter_: Optional[Callable]
    extra_args_: Optional[Dict[str, Any]]
    params_: Optional[Dict[str, Any]]

    def __init__(
        self,
        *,
        incoming_column_name: str,
        incoming_column_is_numeric: bool,
        derived_column_names: Iterable[str],
        treatment: str,
    ):
        """

        :param incoming_column_name:
        :param incoming_column_is_numeric:
        :param derived_column_names:
        :param treatment:
        """

        assert isinstance(incoming_column_name, str)
        assert isinstance(incoming_column_is_numeric, bool)
        assert not isinstance(derived_column_names, str)
        derived_column_names = list(derived_column_names)
        assert len(derived_column_names) > 0
        assert numpy.all([isinstance(dni, str) for dni in derived_column_names])
        assert isinstance(treatment, str)
        self.incoming_column_name_ = incoming_column_name
        self.incoming_column_is_numeric_ = incoming_column_is_numeric
        self.derived_column_names_ = derived_column_names
        self.treatment_ = treatment
        self.need_cross_treatment_ = False
        self.refitter_ = None
        self.extra_args_ = None
        self.params_ = None

    @abc.abstractmethod
    def transform(self, data_frame: pandas.DataFrame) -> pandas.DataFrame:
        """
        return a transformed data frame

        :rtype: pandas.DataFrame
        :param data_frame: incoming values
        :return: transformed values
        """

    @abc.abstractmethod
    def description_matrix(self) -> pandas.DataFrame:
        """
        Return description of transform as a data frame.

        :return: description of transform.
        """


class TreatmentPlan:
    """
    Class to carry treatment plans.
    """

    outcome_name: Optional[str]
    cols_to_copy: Tuple[str, ...]
    num_list: Tuple[str, ...]
    cat_list: Tuple[str, ...]
    xforms: Tuple[VarTransform, ...]

    def __init__(
            self,
            *,
            outcome_name: Optional[str] = None,
            cols_to_copy: Optional[Iterable[str]] = None,
            num_list: Optional[Iterable[str]] = None,
            cat_list: Optional[Iterable[str]] = None,
            xforms: Iterable[Optional[VarTransform]]):
        self.outcome_name = outcome_name
        if cols_to_copy is None:
            self.cols_to_copy = tuple()
        else:
            assert not isinstance(cols_to_copy, str)
            self.cols_to_copy = tuple(cols_to_copy)
        if num_list is None:
            self.num_list = tuple()
        else:
            assert not isinstance(num_list, str)
            self.num_list = tuple(num_list)
        if cat_list is None:
            self.cat_list = tuple()
        else:
            assert not isinstance(cat_list, str)
            self.cat_list = tuple(cat_list)
        non_empty_xforms: List[VarTransform] = [x for x in xforms if x is not None]
        self.xforms = tuple(non_empty_xforms)
        for c in self.cols_to_copy:
            assert isinstance(c, str)
        for c in self.num_list:
            assert isinstance(c, str)
        for c in self.cat_list:
            assert isinstance(c, str)
        if len(self.xforms) < 1:
            raise ValueError("no treatments generated")
        for x in self.xforms:
            assert isinstance(x, VarTransform)


class MappedCodeTransform(VarTransform):
    """Class for transforms that are a dictionary mapping of strings to numeric values"""

    def __init__(
        self,
        *,
        incoming_column_name: str,
        derived_column_name: str,
        treatment: str,
        code_book: pandas.DataFrame,
    ):
        """

        :param incoming_column_name:
        :param derived_column_name:
        :param treatment:
        :param code_book: Pandas dataframe mapping values to impact codes
        """

        VarTransform.__init__(
            self,
            incoming_column_name=incoming_column_name,
            incoming_column_is_numeric=False,
            derived_column_names=[derived_column_name],
            treatment=treatment,
        )
        self.code_book_ = code_book

    def transform(self, data_frame: pandas.DataFrame) -> pandas.DataFrame:
        """
        return a transformed data frame

        :rtype: pandas.DataFrame
        :param data_frame: incoming values
        :return: transformed values
        """

        incoming_column_name = self.incoming_column_name_
        derived_column_name = self.derived_column_names_[0]
        sf = pandas.DataFrame({incoming_column_name: data_frame[incoming_column_name]})
        bad_posns = vtreat.util.is_bad(sf[incoming_column_name])
        sf.loc[bad_posns, incoming_column_name] = bad_sentinel
        res = pandas.merge(
            sf, self.code_book_, on=[self.incoming_column_name_], how="left", sort=False
        )  # ordered by left table rows
        # could also try pandas .map()
        res = res[[derived_column_name]].copy()
        res.loc[vtreat.util.is_bad(res[derived_column_name]), derived_column_name] = 0.0
        return res

    def description_matrix(self) -> pandas.DataFrame:
        """
        Return description of transform as a data frame.

        :return: description of transform.
        """

        description = pandas.DataFrame(
            {
                "treatment_class": "MappedCodeTransform",
                "treatment": self.treatment_,
                "orig_var": self.incoming_column_name_,
                "orig_was_numeric": self.incoming_column_is_numeric_,
                "variable": self.derived_column_names_[0],
                "value": replace_bad_with_sentinel(
                    self.code_book_[self.incoming_column_name_]
                ),
                "replacement": self.code_book_[self.derived_column_names_[0]].copy(),
            }
        )
        return description


class YAwareMappedCodeTransform(MappedCodeTransform):
    """Class for transforms that are a y-aware dictionary mapping of values"""

    def __init__(
        self,
        incoming_column_name: str,
        derived_column_name: str,
        treatment: str,
        code_book: pandas.DataFrame,
        refitter,
        extra_args: Optional[Dict[str, Any]],
        params: Dict[str, Any],
    ):
        """

        :param incoming_column_name: name of incoming column
        :param derived_column_name: name of incoming column
        :param treatment: name of treatment
        :param code_book: pandas data frame mapping values to codes
        :param refitter: function to re-fit
        :param extra_args: extra args for fit_* functions
        :param params: configuration control parameters
        """

        MappedCodeTransform.__init__(
            self,
            incoming_column_name=incoming_column_name,
            derived_column_name=derived_column_name,
            treatment=treatment,
            code_book=code_book,
        )
        self.need_cross_treatment_ = True
        self.refitter_ = refitter
        self.extra_args_ = extra_args
        self.params_ = params


class CleanNumericTransform(VarTransform):
    """Class for numeric column cleaner."""

    def __init__(self,
                 *,
                 incoming_column_name: str,
                 replacement_value: float):
        """

        :param incoming_column_name:
        :param replacement_value:
        """

        VarTransform.__init__(
            self,
            incoming_column_name=incoming_column_name,
            incoming_column_is_numeric=True,
            derived_column_names=[incoming_column_name],
            treatment="clean_copy",
        )
        self.replacement_value_ = replacement_value

    def transform(self, data_frame: pandas.DataFrame) -> pandas.DataFrame:
        """
        return a transformed data frame

        :rtype: pandas.DataFrame
        :param data_frame: incoming values
        :return: transformed values
        """

        col = vtreat.util.safe_to_numeric_array(data_frame[self.incoming_column_name_])
        bad_posns = vtreat.util.is_bad(col)
        col[bad_posns] = self.replacement_value_
        res = pandas.DataFrame({self.derived_column_names_[0]: col})
        return res

    def description_matrix(self) -> pandas.DataFrame:
        """
        Return description of transform as a data frame.

        :return: description of transform.
        """

        description = pandas.DataFrame(
            {
                "treatment_class": ["CleanNumericTransform"],
                "treatment": [self.treatment_],
                "orig_var": [self.incoming_column_name_],
                "orig_was_numeric": [self.incoming_column_is_numeric_],
                "variable": [self.derived_column_names_[0]],
                "value": [bad_sentinel],
                "replacement": [self.replacement_value_],
            }
        )
        return description


class IndicateMissingTransform(VarTransform):
    """Class for missing value indicator."""

    def __init__(self,
                 *,
                 incoming_column_name: str,
                 incoming_column_is_numeric: bool,
                 derived_column_name: str):
        """

        :param incoming_column_name:
        :param derived_column_name:
        """

        VarTransform.__init__(
            self,
            incoming_column_name=incoming_column_name,
            incoming_column_is_numeric=incoming_column_is_numeric,
            derived_column_names=[derived_column_name],
            treatment="missing_indicator",
        )

    def transform(self, data_frame: pandas.DataFrame) -> pandas.DataFrame:
        """
        return a transformed data frame

        :rtype: pandas.DataFrame
        :param data_frame: incoming values
        :return: transformed values
        """

        col = vtreat.util.is_bad(data_frame[self.incoming_column_name_])
        res = pandas.DataFrame({self.derived_column_names_[0]: col + 0.0})
        return res

    def description_matrix(self) -> pandas.DataFrame:
        """
        Return description of transform as a data frame.

        :return: description of transform.
        """
        description = pandas.DataFrame(
            {
                "treatment_class": ["IndicateMissingTransform"],
                "treatment": [self.treatment_],
                "orig_var": [self.incoming_column_name_],
                "orig_was_numeric": [self.incoming_column_is_numeric_],
                "variable": [self.derived_column_names_[0]],
                "value": [bad_sentinel],
                "replacement": [1.0],
            }
        )
        return description


def fit_clean_code(
    *,
    incoming_column_name: str,
    x,
    params: Dict[str, Any],
    imputation_map: Dict[str, Any],
) -> Optional[VarTransform]:
    """
    Fit numeric clean column imputation transform

    :param incoming_column_name: name of column
    :param x: training values for column
    :param params: control parameter dictionary
    :param imputation_map: per-column map to imputation strategies or values
    :return: transform
    """

    if not vtreat.util.numeric_has_range(x):
        return None
    replacement = params["missingness_imputation"]
    try:
        replacement = imputation_map[incoming_column_name]
    except KeyError:
        pass
    if vtreat.util.can_convert_v_to_numeric(replacement):
        replacement_value = 0.0 + replacement
    elif callable(replacement):
        replacement_value = vtreat.util.summarize_column(x, fn=replacement)
    else:
        raise TypeError(
            "unexpected imputation type "
            + str(type(replacement))
            + " ("
            + incoming_column_name
            + ")"
        )
    if (
        pandas.isnull(replacement_value)
        or math.isnan(replacement_value)
        or math.isinf(replacement_value)
    ):
        raise ValueError(
            "replacement was bad "
            + incoming_column_name
            + ": "
            + str(replacement_value)
        )
    return CleanNumericTransform(
        incoming_column_name=incoming_column_name, replacement_value=replacement_value
    )


def fit_regression_impact_code(
    *,
    incoming_column_name: str,
    x,
    y,
    extra_args: Optional[Dict[str, Any]],
    params: Dict[str, Any],
) -> Optional[VarTransform]:
    """
    Fit regression impact code transform

    :param incoming_column_name:
    :param x: training explanatory values
    :param y: training dependent values
    :param extra_args: optional extra arguments for fit_ methods
    :param params: control parameter dictionary
    :return:
    """

    sf = vtreat.util.grouped_by_x_statistics(x, y)
    if sf.shape[0] <= 1:
        return None
    if params["use_hierarchical_estimate"]:
        sf["_impact_code"] = sf["_hest"] - sf["_gm"]
    else:
        sf["_impact_code"] = sf["_group_mean"] - sf["_gm"]
    sf = sf.loc[:, ["x", "_impact_code"]].copy()
    newcol = incoming_column_name + "_impact_code"
    sf.columns = [incoming_column_name, newcol]
    return YAwareMappedCodeTransform(
        incoming_column_name=incoming_column_name,
        derived_column_name=newcol,
        treatment="impact_code",
        code_book=sf,
        refitter=fit_regression_impact_code,
        extra_args=extra_args,
        params=params,
    )


def fit_regression_deviation_code(
    *,
    incoming_column_name: str,
    x,
    y,
    extra_args: Optional[Dict[str, Any]],
    params: Dict[str, Any],
) -> Optional[VarTransform]:
    """
    Fit regression deviation code transform

    :param incoming_column_name:
    :param x: training explanatory values
    :param y: training dependent values
    :param extra_args: optional extra arguments for fit_ methods
    :param params: control parameter dictionary
    :return:
    """

    sf = vtreat.util.grouped_by_x_statistics(x, y)
    if sf.shape[0] <= 1:
        return None
    sf["_deviation_code"] = numpy.sqrt(sf["_var"])
    sf = sf.loc[:, ["x", "_deviation_code"]].copy()
    newcol = incoming_column_name + "_deviation_code"
    sf.columns = [incoming_column_name, newcol]
    return YAwareMappedCodeTransform(
        incoming_column_name=incoming_column_name,
        derived_column_name=newcol,
        treatment="deviation_code",
        code_book=sf,
        refitter=fit_regression_deviation_code,
        extra_args=extra_args,
        params=params,
    )


def fit_binomial_impact_code(
    *,
    incoming_column_name: str,
    x,
    y,
    extra_args: Dict[str, Any],
    params: Dict[str, Any],
) -> Optional[VarTransform]:
    """
    Fit categorical impact code.

    :param incoming_column_name:
    :param x: training explanatory values
    :param y: training dependent values
    :param extra_args: required extra arguments for fit_ methods
    :param params: control parameter dictionary
    :return:
    """
    outcome_target = (
        extra_args["outcome_target"],
    )  # TODO: document why this is a tuple
    var_suffix = extra_args["var_suffix"]
    y = numpy.asarray(numpy.asarray(y) == outcome_target, dtype=float)
    sf = vtreat.util.grouped_by_x_statistics(x, y)
    if sf.shape[0] <= 1:
        return None
    eps = 1.0e-3
    if params["use_hierarchical_estimate"]:
        sf["_logit_code"] = numpy.log((sf["_hest"] + eps) / (sf["_gm"] + eps))
    else:
        sf["_logit_code"] = numpy.log((sf["_group_mean"] + eps) / (sf["_gm"] + eps))
    sf = sf.loc[:, ["x", "_logit_code"]].copy()
    newcol = incoming_column_name + "_logit_code" + var_suffix
    sf.columns = [incoming_column_name, newcol]
    return YAwareMappedCodeTransform(
        incoming_column_name=incoming_column_name,
        derived_column_name=newcol,
        treatment="logit_code",
        code_book=sf,
        refitter=fit_binomial_impact_code,
        extra_args=extra_args,
        params=params,
    )


class IndicatorCodeTransform(VarTransform):
    """Class for indicator codes"""

    def __init__(
        self,
        *,
        incoming_column_name: str,
        derived_column_names: List[str],
        levels: List,
        sparse_indicators: bool = False,
    ):
        """

        :param incoming_column_name:
        :param derived_column_names:
        :param levels: leves we are encoding to indicators
        :param sparse_indicators: if True use sparse data structure
        """
        VarTransform.__init__(
            self,
            incoming_column_name=incoming_column_name,
            incoming_column_is_numeric=False,
            derived_column_names=derived_column_names,
            treatment="indicator_code",
        )
        self.levels_ = levels
        self.sparse_indicators_ = sparse_indicators

    def transform(self, data_frame):
        """
        return a transformed data frame

        :rtype: pandas.DataFrame
        :param data_frame: incoming values
        :return: transformed values
        """

        incoming_column_name = self.incoming_column_name_
        sf = pandas.DataFrame({incoming_column_name: data_frame[incoming_column_name]})
        bad_posns = vtreat.util.is_bad(sf[incoming_column_name])
        sf.loc[bad_posns, incoming_column_name] = bad_sentinel
        col = sf[self.incoming_column_name_]

        def f(i: int):
            """transform one column"""
            v = numpy.asarray(col == self.levels_[i]) + 0.0  # return numeric 0/1 coding
            if self.sparse_indicators_:
                v = pandas.arrays.SparseArray(v, fill_value=0.0)
            return v

        res = None
        for ii in range(len(self.levels_)):
            if res is None:
                res = pandas.DataFrame({self.derived_column_names_[ii]: f(ii)})
            else:
                res[self.derived_column_names_[ii]] = f(ii)
        return res

    def description_matrix(self) -> pandas.DataFrame:
        """
        Return description of transform as a data frame.

        :return: description of transform.
        """

        description = pandas.DataFrame(
            {
                "treatment_class": "IndicatorCodeTransform",
                "treatment": self.treatment_,
                "orig_var": self.incoming_column_name_,
                "orig_was_numeric": self.incoming_column_is_numeric_,
                "variable": self.derived_column_names_.copy(),
                "value": replace_bad_with_sentinel(self.levels_),
                "replacement": 1.0,
            }
        )
        return description


def fit_indicator_code(
    *,
    incoming_column_name: str,
    x,
    min_fraction: float = 0.0,
    max_levels: Optional[int] = None,
    sparse_indicators: bool = False,
) -> Optional[VarTransform]:
    """
    Fit indicator codes

    :param incoming_column_name:
    :param x: training explanatory variables
    :param min_fraction:
    :param max_levels:
    :param sparse_indicators:
    :return:
    """

    sf = pandas.DataFrame({incoming_column_name: x})
    bad_posns = vtreat.util.is_bad(sf[incoming_column_name])
    sf.loc[bad_posns, incoming_column_name] = bad_sentinel
    counts = sf[incoming_column_name].value_counts()
    n = sf.shape[0]
    counts = counts[counts > 0]
    counts = counts[counts >= min_fraction * n]  # no more than 1/min_fraction symbols
    levels = [str(v) for v in counts.index]
    if (max_levels is not None) and (len(levels) > max_levels):
        level_frame = pandas.DataFrame({"levels": levels, "counts": counts})
        level_frame.sort_values(
            by=["counts", "levels"],
            ascending=[False, True],
            inplace=True,
            axis=0,
            ignore_index=True,
        )
        value_vec = level_frame["levels"].values
        levels = [value_vec[i] for i in range(max_levels)]
    if len(levels) < 1:
        return None
    return IndicatorCodeTransform(
        incoming_column_name=incoming_column_name,
        derived_column_names=vtreat.util.build_level_codes(incoming_column_name, levels),
        levels=levels,
        sparse_indicators=sparse_indicators,
    )


def fit_prevalence_code(incoming_column_name: str, x) -> Optional[VarTransform]:
    """
    Fit a prevalence code

    :param incoming_column_name:
    :param x: training explanatory values
    :return:
    """

    sf = pandas.DataFrame({"x": x})
    bad_posns = vtreat.util.is_bad(sf["x"])
    sf.loc[bad_posns, "x"] = bad_sentinel
    sf.reset_index(inplace=True, drop=True)
    n = sf.shape[0]
    sf["_ni"] = 1.0
    sf = pandas.DataFrame(sf.groupby("x")["_ni"].sum())
    sf.reset_index(inplace=True, drop=False)
    sf["_hest"] = sf["_ni"] / n
    sf = sf.loc[:, ["x", "_hest"]].copy()
    newcol = incoming_column_name + "_prevalence_code"
    sf.columns = [incoming_column_name, newcol]
    sf[incoming_column_name] = sf[incoming_column_name].astype(str)
    sf.reset_index(inplace=True, drop=True)
    return MappedCodeTransform(
        incoming_column_name=incoming_column_name,
        derived_column_name=newcol,
        treatment="prevalence_code",
        code_book=sf
    )


# noinspection PyPep8Naming
def _prepare_variable_lists(
    *,
    X,
    cols_to_copy: Optional[Iterable[str]],
    var_list: Optional[Iterable[str]],
) -> Tuple[List[str], List[str], List[str], List[str], List[str]]:
    """
    Prepare lists of variables for variable treatment.

    :param X: dependent variables
    :param cols_to_copy: columns to copy
    :param var_list: dependent variable names, if empty all non outcome and copy columns are used
    :return: cat_list, cols_to_copy, mis_list, num_list, var_list lists
    """
    if var_list is None:
        var_list = list(X.columns)
    else:
        var_list = list(var_list)
    if len(var_list) < 1:
        var_list = list(X.columns)
    assert len(var_list) > 0
    if cols_to_copy is None:
        cols_to_copy = []
    else:
        cols_to_copy = list(cols_to_copy)
    copy_set = set(cols_to_copy)
    var_list = [co for co in var_list if (not (co in copy_set))]
    v_counts = {v: vtreat.util.get_unique_value_count(X[v]) for v in var_list}
    var_list = {v for v in var_list if v_counts[v] > 1}
    if len(var_list) <= 0:
        raise ValueError("no variables")
    n = X.shape[0]
    all_bad: List[str] = []
    mis_list: List[str] = []
    for vi in var_list:
        n_bad = numpy.sum(vtreat.util.is_bad(X[vi]))
        if n_bad >= n:
            all_bad.append(vi)
        if (n_bad > 0) and (n_bad < n):
            mis_list.append(vi)
    var_list = [co for co in var_list if (not (co in set(all_bad)))]
    num_list = [co for co in var_list if vtreat.util.can_convert_v_to_numeric(X[co])]
    cat_list = [co for co in var_list if co not in set(num_list)]
    id_like = [co for co in cat_list if v_counts[co] >= n]
    if len(id_like) > 0:
        warnings.warn(
            "variable(s) "
            + ", ".join(id_like)
            + " have unique values per-row, dropping"
        )
        cat_list = [co for co in var_list if co not in set(id_like)]
    return cat_list, cols_to_copy, mis_list, num_list, var_list


# noinspection PyPep8Naming
def fit_numeric_outcome_treatment(
    *,
    X,
    y,
    var_list: Optional[Iterable[str]],
    outcome_name: str,
    cols_to_copy: Optional[Iterable[str]],
    params: Dict[str, Any],
    imputation_map: Dict[str, Any],
) -> TreatmentPlan:
    """
    Fit set of treatments in a regression situation.

    :param X: training explanatory values
    :param y: training dependent values
    :param var_list: list of dependent variable names, if empty all non outcome and copy columns are used
    :param outcome_name: name for outcome column
    :param cols_to_copy: list of columns to copy to output
    :param params: control parameter dictionary
    :param imputation_map: per-column map to imputation strategies or values
    :return: transform plan
    """
    cat_list, cols_to_copy, mis_list, num_list, var_list = _prepare_variable_lists(
        X=X, cols_to_copy=cols_to_copy, var_list=var_list
    )
    xforms: List[Optional[VarTransform]] = []
    if "missing_indicator" in params["coders"]:
        num_set = set(num_list)
        for vi in mis_list:
            xforms.append(
                IndicateMissingTransform(
                    incoming_column_name=vi,
                    incoming_column_is_numeric=vi in num_set,
                    derived_column_name=vi + "_is_bad"
                )
            )
    if "clean_copy" in params["coders"]:
        for vi in num_list:
            xform = fit_clean_code(
                incoming_column_name=vi,
                x=X[vi],
                params=params,
                imputation_map=imputation_map,
            )
            if xform is not None:
                xforms.append(xform)
    for vi in cat_list:
        if "impact_code" in params["coders"]:
            xforms.append(
                fit_regression_impact_code(
                    incoming_column_name=vi,
                    x=numpy.asarray(X[vi]),
                    y=y,
                    extra_args=None,
                    params=params,
                )
            )
        if "deviation_code" in params["coders"]:
            xforms.append(
                fit_regression_deviation_code(
                    incoming_column_name=vi,
                    x=numpy.asarray(X[vi]),
                    y=y,
                    extra_args=None,
                    params=params,
                )
            )
        if "prevalence_code" in params["coders"]:
            xforms.append(
                fit_prevalence_code(incoming_column_name=vi, x=numpy.asarray(X[vi]))
            )
        if "indicator_code" in params["coders"]:
            xforms.append(
                fit_indicator_code(
                    incoming_column_name=vi,
                    x=numpy.asarray(X[vi]),
                    min_fraction=params["indicator_min_fraction"],
                    sparse_indicators=params["sparse_indicators"],
                )
            )
    for stp in params["user_transforms"]:
        stp.fit(X=X[var_list], y=y)
    return TreatmentPlan(
        outcome_name=outcome_name,
        cols_to_copy=cols_to_copy,
        num_list=num_list,
        cat_list=cat_list,
        xforms=xforms,
    )


# noinspection PyPep8Naming
def fit_binomial_outcome_treatment(
    *,
    X,
    y,
    outcome_target,
    var_list: Optional[Iterable[str]],
    outcome_name: str,
    cols_to_copy: Optional[Iterable[str]],
    params: Dict[str, Any],
    imputation_map: Dict[str, Any],
) -> TreatmentPlan:
    """

    :param X: training explanatory values
    :param y: training dependent values
    :param outcome_target: dependent value to consider positive or in class
    :param var_list: list of variables to process, if empty all non outcome and copy columns are used
    :param outcome_name: name for outcome column
    :param cols_to_copy: list of columns to copy to output
    :param params: control parameter dictionary
    :param imputation_map: per-column map to imputation strategies or values
    :return: transform plan
    """
    cat_list, cols_to_copy, mis_list, num_list, var_list = _prepare_variable_lists(
        X=X, cols_to_copy=cols_to_copy, var_list=var_list
    )
    xforms: List[Optional[VarTransform]] = []
    if "missing_indicator" in params["coders"]:
        num_set = set(num_list)
        for vi in mis_list:
            xforms.append(
                IndicateMissingTransform(
                    incoming_column_name=vi,
                    incoming_column_is_numeric=vi in num_set,
                    derived_column_name=vi + "_is_bad"
                )
            )
    if "clean_copy" in params["coders"]:
        for vi in num_list:
            xform = fit_clean_code(
                incoming_column_name=vi,
                x=X[vi],
                params=params,
                imputation_map=imputation_map,
            )
            if xform is not None:
                xforms.append(xform)
    extra_args = {"outcome_target": outcome_target, "var_suffix": ""}
    for vi in cat_list:
        if "logit_code" in params["coders"]:
            xforms.append(
                fit_binomial_impact_code(
                    incoming_column_name=vi,
                    x=numpy.asarray(X[vi]),
                    y=y,
                    extra_args=extra_args,
                    params=params,
                )
            )
        if "prevalence_code" in params["coders"]:
            xforms.append(
                fit_prevalence_code(incoming_column_name=vi, x=numpy.asarray(X[vi]))
            )
        if "indicator_code" in params["coders"]:
            xforms.append(
                fit_indicator_code(
                    incoming_column_name=vi,
                    x=numpy.asarray(X[vi]),
                    min_fraction=params["indicator_min_fraction"],
                    sparse_indicators=params["sparse_indicators"],
                )
            )
    for stp in params["user_transforms"]:
        stp.fit(X=X[var_list], y=y)
    return TreatmentPlan(
        outcome_name=outcome_name,
        cols_to_copy=cols_to_copy,
        num_list=num_list,
        cat_list=cat_list,
        xforms=xforms,
    )


# noinspection PyPep8Naming
def fit_multinomial_outcome_treatment(
    *,
    X,
    y,
    var_list: Optional[Iterable[str]],
    outcome_name: str,
    cols_to_copy: Optional[Iterable[str]],
    params: Dict[str, Any],
    imputation_map: Dict[str, Any],
) -> TreatmentPlan:
    """
    Fit a variable treatment for multinomial outcomes.

    :param X: training explanatory values
    :param y: training dependent values
    :param var_list: list of variables to process, if empty all non outcome and copy columns are used
    :param outcome_name: name for outcome column
    :param cols_to_copy: list of columns to copy to output
    :param params: control parameter dictionary
    :param imputation_map: per-column map to imputation strategies or values
    :return:
    """

    outcomes = [oi for oi in set(y)]
    assert len(outcomes) > 1
    cat_list, cols_to_copy, mis_list, num_list, var_list = _prepare_variable_lists(
        X=X, cols_to_copy=cols_to_copy, var_list=var_list
    )
    xforms: List[Optional[VarTransform]] = []
    if "missing_indicator" in params["coders"]:
        num_set = set(num_list)
        for vi in mis_list:
            xforms.append(
                IndicateMissingTransform(
                    incoming_column_name=vi,
                    incoming_column_is_numeric=vi in num_set,
                    derived_column_name=vi + "_is_bad"
                )
            )
    if "clean_copy" in params["coders"]:
        for vi in num_list:
            xform = fit_clean_code(
                incoming_column_name=vi,
                x=X[vi],
                params=params,
                imputation_map=imputation_map,
            )
            if xform is not None:
                xforms.append(xform)
    for vi in cat_list:
        for outcome in outcomes:
            if "impact_code" in params["coders"]:
                extra_args = {
                    "outcome_target": outcome,
                    "var_suffix": ("_" + str(outcome)),
                }
                xforms.append(
                    fit_binomial_impact_code(
                        incoming_column_name=vi,
                        x=numpy.asarray(X[vi]),
                        y=y,
                        extra_args=extra_args,
                        params=params,
                    )
                )
        if "prevalence_code" in params["coders"]:
            xforms.append(
                fit_prevalence_code(incoming_column_name=vi, x=numpy.asarray(X[vi]))
            )
        if "indicator_code" in params["coders"]:
            xforms.append(
                fit_indicator_code(
                    incoming_column_name=vi,
                    x=numpy.asarray(X[vi]),
                    min_fraction=params["indicator_min_fraction"],
                    sparse_indicators=params["sparse_indicators"],
                )
            )
    for stp in params["user_transforms"]:
        stp.fit(X=X[var_list], y=y)
    return TreatmentPlan(
        outcome_name=outcome_name,
        cols_to_copy=cols_to_copy,
        num_list=num_list,
        cat_list=cat_list,
        xforms=xforms,
    )


# noinspection PyPep8Naming
def fit_unsupervised_treatment(
    *,
    X,
    var_list: Optional[Iterable[str]],
    outcome_name: str,
    cols_to_copy: Optional[Iterable[str]],
    params: Dict[str, Any],
    imputation_map: Dict[str, Any],
) -> TreatmentPlan:
    """
    Fit a data treatment in the unsupervised case.

    :param X: training explanatory values
    :param var_list: list of variables to process, if empty all non copy columns are used
    :param outcome_name: name for outcome column
    :param cols_to_copy: list of columns to copy to output
    :param params: control parameter dictionary
    :param imputation_map: per-column map to imputation strategies or values
    :return:
    """

    cat_list, cols_to_copy, mis_list, num_list, var_list = _prepare_variable_lists(
        X=X, cols_to_copy=cols_to_copy, var_list=var_list
    )
    xforms: List[Optional[VarTransform]] = []
    if "missing_indicator" in params["coders"]:
        num_set = set(num_list)
        for vi in mis_list:
            xforms.append(
                IndicateMissingTransform(
                    incoming_column_name=vi,
                    incoming_column_is_numeric=vi in num_set,
                    derived_column_name=vi + "_is_bad"
                )
            )
    if "clean_copy" in params["coders"]:
        for vi in num_list:
            xform = fit_clean_code(
                incoming_column_name=vi,
                x=X[vi],
                params=params,
                imputation_map=imputation_map,
            )
            if xform is not None:
                xforms.append(xform)
    for vi in cat_list:
        if "prevalence_code" in params["coders"]:
            xforms.append(
                fit_prevalence_code(incoming_column_name=vi, x=numpy.asarray(X[vi]))
            )
        if "indicator_code" in params["coders"]:
            xforms.append(
                fit_indicator_code(
                    incoming_column_name=vi,
                    x=numpy.asarray(X[vi]),
                    min_fraction=params["indicator_min_fraction"],
                    max_levels=params["indicator_max_levels"],
                    sparse_indicators=params["sparse_indicators"],
                )
            )
    for stp in params["user_transforms"]:
        stp.fit(X=X[var_list], y=None)
    return TreatmentPlan(
        outcome_name=outcome_name,
        cols_to_copy=cols_to_copy,
        num_list=num_list,
        cat_list=cat_list,
        xforms=xforms,
    )


def pre_prep_frame(
    x: pandas.DataFrame,
    *,
    col_list: Optional[Iterable[str]],
    cols_to_copy: Optional[Iterable[str]],
    cat_cols: Optional[Iterable[str]] = None,
) -> pandas.DataFrame:
    """
    Create a copy of pandas.DataFrame x restricted to col_list union cols_to_copy with col_list - cols_to_copy
    converted to only string and numeric types.  New pandas.DataFrame has trivial indexing.  If col_list
    is empty it is interpreted as all columns.

    :param x:
    :param col_list:
    :param cols_to_copy:
    :param cat_cols:
    :return:
    """

    if cols_to_copy is None:
        cols_to_copy = []
    else:
        cols_to_copy = list(cols_to_copy)
    if col_list is None:
        col_list = []
    else:
        col_list = list(col_list)
    if len(col_list) <= 0:
        col_list = list(x.columns)
    x_set = set(x.columns)
    col_set = set(col_list)
    for ci in cols_to_copy:
        if (ci in x_set) and (ci not in col_set):
            col_list.append(ci)
    col_set = set(col_list)
    missing_cols = col_set - x_set
    if len(missing_cols) > 0:
        raise KeyError("referred to not-present columns " + str(missing_cols))
    cset = set(cols_to_copy)
    if len(col_list) <= 0:
        raise ValueError("no variables")
    x = x.loc[:, col_list]
    x = x.reset_index(inplace=False, drop=True)
    cat_col_set = None
    if cat_cols is not None:
        cat_col_set = set(cat_cols)
    for c in x.columns:
        if c in cset:
            continue
        bad_ind = vtreat.util.is_bad(x[c])
        if cat_col_set is not None:
            numeric_col = c not in cat_col_set
        else:
            numeric_col = vtreat.util.can_convert_v_to_numeric(x[c])
        if numeric_col:
            x[c] = vtreat.util.safe_to_numeric_array(x[c])
        else:
            # https://stackoverflow.com/questions/22231592/pandas-change-data-type-of-series-to-string
            x[c] = x[c].astype(str)
        x.loc[bad_ind, c] = None
    return x


def _mean_of_single_column_pandas_list(val_list: Iterable[pandas.DataFrame]) -> float:
    """
    Compute the mean of non-nan positions of a bunch of single column data frames

    :param val_list: a list of single column Pandas data frames
    :return: mean, or numpy.nan if there are no values
    """

    if val_list is None:
        return numpy.nan
    val_list = [v for v in val_list]
    if len(val_list) <= 0:
        return numpy.nan
    if len(val_list) <= 1:
        d = val_list[0]
    else:
        d = pandas.concat(val_list, axis=0, sort=False)
    col = d.columns[0]
    d = d.loc[numpy.logical_not(vtreat.util.is_bad(d[col])), [col]]
    if d.shape[0] < 1:
        return numpy.nan
    res = numpy.mean(d[col].values)
    assert isinstance(res, float)  # type hint for PyCharm IDE
    return res


def cross_patch_refit_y_aware_cols(
    *, x: pandas.DataFrame, y, res: pandas.DataFrame, plan: TreatmentPlan, cross_plan
) -> None:
    """
    Re fit the y-aware columns according to cross plan.
    Clears out refitter_ values to None.
    Assumes each y-aware variable produces one derived column.

    :param x: explanatory values
    :param y: dependent values
    :param res: transformed frame to patch results into, altered
    :param plan: fitting plan
    :param cross_plan: cross validation plan
    :return: no return, res is altered in place
    """

    if cross_plan is None or len(cross_plan) <= 1:
        for xf in plan.xforms:
            xf.refitter_ = None
        return
    incoming_colset = set(x.columns)
    derived_colset = set(res.columns)
    for xf in plan.xforms:
        if not xf.need_cross_treatment_:
            continue
        incoming_column_name = xf.incoming_column_name_
        derived_column_name = xf.derived_column_names_[0]
        if derived_column_name not in derived_colset:
            continue
        if incoming_column_name not in incoming_colset:
            raise KeyError("missing required column " + incoming_column_name)
        if xf.refitter_ is None:
            raise ValueError(
                "refitter is None: "
                + incoming_column_name
                + " -> "
                + derived_column_name
            )

        # noinspection PyPep8Naming
        def maybe_transform(*, fit, X):
            """conditional patching"""
            if fit is None:
                return None
            return fit.transform(X)

        patches = [
            maybe_transform(
                fit=xf.refitter_(
                    incoming_column_name=incoming_column_name,
                    x=x[incoming_column_name][cp["train"]],
                    y=y[cp["train"]],
                    extra_args=xf.extra_args_,
                    params=xf.params_,
                ),
                X=x.loc[cp["app"], [incoming_column_name]],
            )
            for cp in cross_plan
        ]
        # replace any missing sections with global average (slight data leak potential)
        avg = _mean_of_single_column_pandas_list(
            [pi for pi in patches if pi is not None]
        )
        if numpy.isnan(avg):
            avg = 0
        res[derived_column_name] = avg
        for i in range(len(cross_plan)):
            pi = patches[i]
            if pi is None:
                continue
            pi.reset_index(inplace=True, drop=True)
            cp = cross_plan[i]
            res.loc[cp["app"], derived_column_name] = numpy.asarray(
                pi[derived_column_name]
            ).reshape((len(pi),))
        res.loc[vtreat.util.is_bad(res[derived_column_name]), derived_column_name] = avg
    for xf in plan.xforms:
        xf.refitter_ = None


def cross_patch_user_y_aware_cols(
    *, x: pandas.DataFrame, y, res: pandas.DataFrame, params: Dict[str, Any], cross_plan
) -> None:
    """
    Re fit the user y-aware columns according to cross plan.
    Assumes each y-aware variable produces one derived column.

    :param x: explanatory values
    :param y: dependent values
    :param res: transformed frame to patch results into, altered
    :param params: control parameter dictionary
    :param cross_plan: cross validation plan
    :return: no return, res altered in place
    """
    if cross_plan is None or len(cross_plan) <= 1:
        return
    incoming_colset = set(x.columns)
    derived_colset = set(res.columns)
    if len(derived_colset) <= 0:
        return
    for ut in params["user_transforms"]:
        if not ut.y_aware_:
            continue
        instersect_in = incoming_colset.intersection(set(ut.incoming_vars_))
        instersect_out = derived_colset.intersection(set(ut.derived_vars_))
        if len(instersect_out) <= 0:
            continue
        if len(instersect_out) != len(ut.derived_vars_):
            raise ValueError("not all derived columns are in res frame")
        if len(instersect_in) != len(ut.incoming_vars_):
            raise KeyError("missing required columns")
        patches = [
            ut.fit(X=x.loc[cp["train"], ut.incoming_vars_], y=y[cp["train"]]).transform(
                X=x.loc[cp["app"], ut.incoming_vars_]
            )
            for cp in cross_plan
        ]
        for col in ut.derived_vars_:
            # replace any missing sections with global average (slight data leak potential)
            avg = _mean_of_single_column_pandas_list(
                [pi.loc[:, [col]] for pi in patches if pi is not None]
            )
            if numpy.isnan(avg):
                avg = 0
            res[col] = avg
            for i in range(len(cross_plan)):
                pi = patches[i]
                if pi is None:
                    continue
                pi.reset_index(inplace=True, drop=True)
                cp = cross_plan[i]
                res.loc[cp["app"], col] = numpy.asarray(pi[col]).reshape((len(pi),))
            res.loc[vtreat.util.is_bad(res[col]), col] = avg


def score_plan_variables(
    cross_frame: pandas.DataFrame,
    outcome,
    plan: TreatmentPlan,
    params: Dict[str, Any],
    *,
    is_classification: bool = False,
) -> pandas.DataFrame:
    """
    Quality score variables to build up score frame.

    :param cross_frame: cross transformed explanatory variables
    :param outcome: dependent variable
    :param plan: treatment plan
    :param params: control parameter dictionary
    :param is_classification: logical, if True classification if False regression
    :return: score frame
    """

    def describe_xf(xf):
        """describe variable transform"""
        description = pandas.DataFrame({"variable": xf.derived_column_names_})
        description["orig_variable"] = xf.incoming_column_name_
        description["treatment"] = xf.treatment_
        description["y_aware"] = xf.need_cross_treatment_
        return description

    def describe_ut(ut):
        """describe user variable transform"""
        description = pandas.DataFrame(
            {"orig_variable": ut.incoming_vars_, "variable": ut.derived_vars_}
        )
        description["treatment"] = ut.treatment_
        description["y_aware"] = ut.y_aware_
        return description

    var_table = pandas.concat(
        [describe_xf(xf) for xf in plan.xforms]
        + [
            describe_ut(ut)
            for ut in params["user_transforms"]
            if len(ut.incoming_vars_) > 0
        ],
        sort=False,
    )
    var_table.reset_index(inplace=True, drop=True)
    sf = vtreat.util.score_variables(
        cross_frame,
        variables=var_table["variable"],
        outcome=outcome,
        is_classification=is_classification,
    )
    score_frame = pandas.merge(var_table, sf, how="left", on=["variable"], sort=False)
    num_treatment_types = len(score_frame["treatment"].unique())
    score_frame["_one"] = 1.0
    score_frame["vcount"] = score_frame.groupby("treatment")["_one"].transform("sum")
    score_frame["default_threshold"] = 1.0 / (
        score_frame["vcount"] * num_treatment_types
    )
    score_frame.drop(["_one"], axis=1, inplace=True)
    score_frame["recommended"] = numpy.logical_and(
        score_frame["has_range"],
        numpy.logical_and(
            numpy.logical_not(
                numpy.logical_or(
                    numpy.isnan(score_frame["significance"]),
                    numpy.isnan(score_frame["PearsonR"]),
                )
            ),
            numpy.logical_and(
                score_frame["significance"] < score_frame["default_threshold"],
                numpy.logical_or(
                    score_frame["PearsonR"] > 0.0,
                    numpy.logical_not(score_frame["y_aware"]),
                ),
            ),
        ),
    )
    return score_frame


def pseudo_score_plan_variables(
    *, cross_frame, plan: TreatmentPlan, params: Dict[str, Any]
) -> pandas.DataFrame:
    """
    Build a score frame look-alike for unsupervised case.

    :param cross_frame: cross transformed explanatory variables
    :param plan: treatment plan
    :param params: control parameter dictionary
    :return: score frame
    """

    def describe_xf(xf):
        """describe variable transform"""
        description = pandas.DataFrame({"variable": xf.derived_column_names_})
        description["orig_variable"] = xf.incoming_column_name_
        description["treatment"] = xf.treatment_
        description["y_aware"] = xf.need_cross_treatment_
        return description

    def describe_ut(ut):
        """describe user variable transform"""
        description = pandas.DataFrame(
            {"orig_variable": ut.incoming_vars_, "variable": ut.derived_vars_}
        )
        description["treatment"] = ut.treatment_
        description["y_aware"] = ut.y_aware_
        return description

    score_frame = pandas.concat(
        [describe_xf(xf) for xf in plan.xforms]
        + [
            describe_ut(ut)
            for ut in params["user_transforms"]
            if len(ut.incoming_vars_) > 0
        ],
        sort=False,
    )
    score_frame.reset_index(inplace=True, drop=True)

    score_frame["has_range"] = [
        vtreat.util.numeric_has_range(cross_frame[c]) for c in score_frame["variable"]
    ]
    score_frame["PearsonR"] = numpy.nan
    score_frame["significance"] = numpy.nan
    score_frame["recommended"] = score_frame["has_range"].copy()
    score_frame["_one"] = 1.0
    score_frame["vcount"] = score_frame.groupby("treatment")["_one"].transform("sum")
    score_frame.drop(["_one"], axis=1, inplace=True)
    return score_frame


class VariableTreatment(abc.ABC, sklearn.base.BaseEstimator, sklearn.base.TransformerMixin):
    """
    Class for variable treatments, implements much of the sklearn pipeline/transformer
    API. https://sklearn-template.readthedocs.io/en/latest/user_guide.html#transformer
    """

    result_restriction: Optional[Set[str]]
    outcome_name_: Optional[str]
    outcome_target_: Optional[Any]
    var_list_: List[str]
    cols_to_copy_: List[str]
    params_: Dict[str, Any]
    plan_: Optional[TreatmentPlan]
    score_frame_: Optional[pandas.DataFrame]
    imputation_map_: Dict[str, Callable]
    last_fit_x_id_: Optional[str]
    cross_plan_: Optional[List[Dict[str, List[int]]]]
    cross_rows_: Optional[int]
    last_result_columns: Optional[List[str]]

    def __init__(
        self,
        *,
        var_list: Optional[Iterable[str]] = None,
        outcome_name: Optional[str] = None,
        outcome_target: Optional[Any] = None,
        cols_to_copy: Optional[Iterable[str]] = None,
        params: Optional[Dict[str, Any]] = None,
        imputation_map: Optional[Dict[str, Any]] = None,
    ):
        """

        :param var_list: variables we intend to encode, empty means all
        :param outcome_name: column name of outcome
        :param outcome_target: outcome column value we consider in class or True
        :param cols_to_copy: columns to not process, but copy over
        :param params: control and configuration parameters
        :param imputation_map: per column imputation strategies or values
        """
        if var_list is None:
            var_list = []
        else:
            var_list = vtreat.util.unique_items_in_order(var_list)
        if cols_to_copy is None:
            cols_to_copy = []
        else:
            cols_to_copy = vtreat.util.unique_items_in_order(cols_to_copy)
        if (outcome_name is not None) and (outcome_name not in set(cols_to_copy)):
            cols_to_copy.append(outcome_name)
        confused = set(cols_to_copy).intersection(set(var_list))
        if len(confused) > 0:
            raise ValueError(
                "variables in treatment plan and non-treatment: " + ", ".join(confused)
            )
        if imputation_map is None:
            imputation_map = dict()
        self.outcome_name_ = outcome_name
        self.outcome_target_ = outcome_target
        self.var_list_ = [vi for vi in var_list if vi not in set(cols_to_copy)]
        self.cols_to_copy_ = cols_to_copy
        if params is not None:
            self.params_ = params.copy()
        else:
            self.params_ = dict()
        self.imputation_map_ = imputation_map.copy()
        self.plan_ = None
        self.score_frame_ = None
        self.cross_rows_ = None
        self.cross_plan_ = None
        self.last_fit_x_id_ = None
        self.last_result_columns = None
        self.result_restriction = None
        self.clear()

    def check_column_names(self, col_names: Iterable[str]) -> None:
        """
        Check that none of the column names we are working with are non-unique.
        Also check variable columns are all present (columns to copy and outcome allowed to be missing).

        :param col_names:
        :return: None, raises exception if there is a problem
        """

        col_names = [c for c in col_names]
        to_check = set(self.var_list_)
        if self.outcome_name_ is not None:
            to_check.add(self.outcome_name_)
        if self.cols_to_copy_ is not None:
            to_check.update(self.cols_to_copy_)
        seen = [c for c in col_names if c in to_check]
        if len(seen) != len(set(seen)):
            raise ValueError("duplicate column names in frame")
        missing = set(self.var_list_) - set(col_names)
        if len(missing) > 0:
            raise ValueError(f"missing required columns: {missing}")

    def clear(self) -> None:
        """reset state"""
        self.plan_ = None
        self.score_frame_ = None
        self.cross_rows_ = None
        self.cross_plan_ = None
        self.last_fit_x_id_ = None
        self.last_result_columns = None
        self.result_restriction = None

    def get_result_restriction(self):
        """accessor"""
        if self.result_restriction is None:
            return None
        return self.result_restriction.copy()

    def set_result_restriction(self, new_vars) -> None:
        """setter"""
        self.result_restriction = None
        if (new_vars is not None) and (len(new_vars) > 0):
            self.result_restriction = set(new_vars)

    @abc.abstractmethod
    def merge_params(self, p: Optional[Dict[str, Any]]) -> Dict[str, Any]:
        """merge in use parameters"""

    # sklearn pipeline step methods

    # https://scikit-learn.org/stable/modules/generated/sklearn.base.TransformerMixin.html

    # noinspection PyPep8Naming, PyUnusedLocal
    def fit(self, X, y=None, **fit_params):
        """
        sklearn fit.

        :param X: explanatory variables
        :param y: (optional) dependent variable
        :param fit_params:
        :return: self (for method chaining)
        """

        self.fit_transform(X=X, y=y)
        return self

    # noinspection PyPep8Naming, PyUnusedLocal
    @abc.abstractmethod
    def fit_transform(self, X, y=None, **fit_params):
        """
        sklearn fit_transform, correct way to trigger cross methods.

        :param X: explanatory variables
        :param y: (optional) dependent variable
        :param fit_params:
        :return: transformed data
        """

    # noinspection PyPep8Naming
    @abc.abstractmethod
    def transform(self, X):
        """
        sklearn transform

        :param X: explanatory variables
        :return: transformed data
        """

    # https://scikit-learn.org/stable/modules/generated/sklearn.base.BaseEstimator.html

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def get_params(self, deep=True):
        """
        vtreat exposes a subset of controls as tunable parameters, users can choose this set
        by specifying the tunable_params list in object construction parameters

        :param deep: ignored
        :return: dict of tunable parameters
        """
        return {ti: self.params_[ti] for ti in self.params_["tunable_params"]}

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def set_params(self, **params):
        """
        vtreat exposes a subset of controls as tunable parameters, users can choose this set
        by specifying the tunable_params list in object construction parameters

        :param params:
        :return: self (for method chaining)
        """

        for (k, v) in params.items():
            if k in self.params_["tunable_params"]:
                self.params_[k] = v
        return self

    # extra methods to look more like sklearn objects

    # https://scikit-learn.org/stable/modules/generated/sklearn.pipeline.Pipeline.html

    # noinspection PyPep8Naming
    def fit_predict(self, X, y=None, **fit_params):
        """
        Alias for fit_transform()

        :param X: explanatory variables
        :param y: (optional) dependent variable
        :param fit_params:
        :return: transformed data
        """
        return self.fit_transform(X=X, y=y, **fit_params)

    # noinspection PyPep8Naming
    def predict(self, X):
        """
        Alias for transform.

        :param X: explanatory variables
        :return: transformed data
        """

        return self.transform(X)

    # noinspection PyPep8Naming
    def predict_proba(self, X):
        """
        Alias for transform.

        :param X: explanatory variables
        :return: transformed data
        """

        return self.transform(X)

    # https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/compose/_column_transformer.py

    def get_feature_names(self, input_features=None):
        """
        Get list of produced feature names.

        :param input_features: Optional, restrict to these features
        :return:
        """
        if self.score_frame_ is None:
            raise ValueError(
                "get_feature_names called on uninitialized vtreat transform"
            )
        if input_features is not None:
            input_features = [c for c in input_features]
        filter_to_recommended = False
        try:
            filter_to_recommended = self.params_["filter_to_recommended"]
        except KeyError:
            pass
        if filter_to_recommended:
            new_vars = [
                self.score_frame_["variable"][i]
                for i in range(self.score_frame_.shape[0])
                if self.score_frame_["has_range"][i]
                and self.score_frame_["recommended"][i]
                and (
                    input_features is None
                    or self.score_frame_["orig_variable"][i] in input_features
                )
            ]
        else:
            new_vars = [
                self.score_frame_["variable"][i]
                for i in range(self.score_frame_.shape[0])
                if self.score_frame_["has_range"][i]
                and (
                    input_features is None
                    or self.score_frame_["orig_variable"][i] in input_features
                )
            ]
        new_vars = new_vars + self.cols_to_copy_
        return new_vars

    def description_matrix(self) -> pandas.DataFrame:
        """
        Return description of transform as a data frame. Does not encode user steps. Not yet implemented for
        multinomial dependent variables.

        :return: description of transform.
        """

        assert self.plan_ is not None  # type hint
        assert self.score_frame_ is not None  # type hint
        xform_steps = [xfi for xfi in self.plan_.xforms]
        frames = [xfi.description_matrix() for xfi in xform_steps]
        res = pandas.concat(frames).reset_index(inplace=False, drop=True)
        # restrict down to non-constant variables
        scored_vars = set(self.score_frame_["variable"][self.score_frame_["has_range"]])
        usable = [v in scored_vars for v in res["variable"]]
        res = res.loc[usable, :].reset_index(inplace=False, drop=True)
        return res


def perform_transform(
    *, x: pandas.DataFrame, transform: VariableTreatment, params: Dict[str, Any]
) -> pandas.DataFrame:
    """
    Transform a data frame.

    :param x: data to be transformed.
    :param transform: transform
    :param params: control parameter dictionary
    :return: new data frame
    """
    assert transform.plan_ is not None  # type hint
    xform_steps = [xfi for xfi in transform.plan_.xforms]
    user_steps = [stp for stp in params["user_transforms"]]
    # restrict down to to results we are going to use
    if (transform.result_restriction is not None) and (
        len(transform.result_restriction) > 0
    ):
        xform_steps = [
            xfi
            for xfi in xform_steps
            if len(
                set(xfi.derived_column_names_).intersection(
                    transform.result_restriction
                )
            )
            > 0
        ]
        user_steps = [
            stp
            for stp in user_steps
            if len(set(stp.derived_vars_).intersection(transform.result_restriction))
            > 0
        ]
    # check all required columns are present
    needs = set()
    for xfi in xform_steps:
        if xfi.incoming_column_name_ is not None:
            needs.add(xfi.incoming_column_name_)
    for stp in user_steps:
        if stp.incoming_vars_ is not None:
            needs.update(stp.incoming_vars_)
    missing = needs - set(x.columns)
    if len(missing) > 0:
        raise ValueError("missing required input columns " + str(missing))
    # do the work
    new_frames = [xfi.transform(x) for xfi in (xform_steps + user_steps)]
    new_frames = [frm for frm in new_frames if (frm is not None) and (frm.shape[1] > 0)]
    # see if we want to copy over any columns
    copy_set = set(transform.plan_.cols_to_copy)
    to_copy = [ci for ci in x.columns if ci in copy_set]
    if len(to_copy) > 0:
        cp = x.loc[:, to_copy].copy()
        new_frames = [cp] + new_frames
    if len(new_frames) <= 0:
        raise ValueError("no columns transformed")
    res = pandas.concat(new_frames, axis=1, sort=False)
    res.reset_index(inplace=True, drop=True)
    return res


def limit_to_appropriate_columns(
    *, res: pandas.DataFrame, transform: VariableTreatment
) -> pandas.DataFrame:
    """
    Limit down to appropriate columns.

    :param res:
    :param transform:
    :return:
    """
    assert transform.plan_ is not None  # type hint
    assert transform.score_frame_ is not None  # type hint
    to_copy = set(transform.plan_.cols_to_copy)
    to_take = set(
        [
            ci
            for ci in transform.score_frame_["variable"][
                transform.score_frame_["has_range"]
            ]
        ]
    )
    if (transform.result_restriction is not None) and (
        len(transform.result_restriction) > 0
    ):
        to_take = to_take.intersection(transform.result_restriction)
    cols_to_keep = [ci for ci in res.columns if (ci in to_copy) or (ci in to_take)]
    if len(cols_to_keep) <= 0:
        raise ValueError("no columns retained")
    res = res[cols_to_keep].copy()
    res.reset_index(inplace=True, drop=True)
    return res



"""
Define user visible vtreat API.
"""

import warnings
from typing import Any, Dict, Iterable, Optional

import pandas
import numpy

import vtreat.vtreat_impl as vtreat_impl
import vtreat.util
import vtreat.cross_plan


def vtreat_parameters(user_params: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    """
    build a vtreat parameters dictionary, adding in user choices

    :param user_params: dictionary of user choices
    :return: dictionary of user and default choices
    """

    params = {
        "use_hierarchical_estimate": True,
        "coders": {
            "clean_copy",
            "missing_indicator",
            "indicator_code",
            "impact_code",
            "deviation_code",
            "logit_code",
            "prevalence_code",
        },
        "filter_to_recommended": True,
        "indicator_min_fraction": 0.02,
        "cross_validation_plan": vtreat.cross_plan.KWayCrossPlanYStratified(),
        "cross_validation_k": 5,
        "user_transforms": [],
        "sparse_indicators": False,
        "missingness_imputation": numpy.mean,
        "check_for_duplicate_frames": True,
        "error_on_duplicate_frames": False,
        "retain_cross_plan": True,
        "tunable_params": ["indicator_min_fraction"],
    }
    pkeys = set(params.keys())
    if user_params is not None:
        for k in user_params.keys():
            if k not in pkeys:
                raise KeyError("parameter key " + str(k) + " not recognized")
            params[k] = user_params[k]
    if params["error_on_duplicate_frames"]:
        params["check_for_duplicate_frames"] = True
    for k in params["tunable_params"]:
        if k not in pkeys:
            raise KeyError("tunable_params key " + str(k) + " not recognized")
    return params


def unsupervised_parameters(
    user_params: Optional[Dict[str, Any]] = None
) -> Dict[str, Any]:
    """
    build a vtreat parameters dictionary for unsupervised tasks, adding in user choices

    :param user_params: dictionary of user choices
    :return: dictionary of user and default choices
    """

    params = {
        "coders": {
            "clean_copy",
            "missing_indicator",
            "indicator_code",
            "prevalence_code",
        },
        "indicator_min_fraction": 0.0,
        "indicator_max_levels": 1000,
        "user_transforms": [],
        "sparse_indicators": False,
        "missingness_imputation": numpy.mean,
        "tunable_params": ["indicator_min_fraction", "indicator_max_levels"],
    }
    pkeys = set(params.keys())
    if user_params is not None:
        for k in user_params.keys():
            if k not in pkeys:
                raise KeyError("parameter key " + str(k) + " not recognized")
            params[k] = user_params[k]
    for k in params["tunable_params"]:
        if k not in pkeys:
            raise KeyError("tunable_params key " + str(k) + " not recognized")
    return params


class NumericOutcomeTreatment(vtreat_impl.VariableTreatment):
    """manage a treatment plan for a numeric outcome (regression)"""

    def __init__(
        self,
        *,
        var_list: Optional[Iterable[str]] = None,
        outcome_name: Optional[str] = None,
        cols_to_copy: Optional[Iterable[str]] = None,
        params: Optional[Dict[str, Any]] = None,
        imputation_map: Optional[Dict[str, Any]] = None,
    ):
        """

         :param var_list: list or tuple of column names, if empty all non outcome and copy columns are used
         :param outcome_name: name of column containing dependent variable
         :param cols_to_copy: list or tuple of column names
         :param params: vtreat.vtreat_parameters()
         :param imputation_map: map of column names to custom missing imputation values or functions
        """
        params = self.merge_params(params)
        vtreat_impl.VariableTreatment.__init__(
            self,
            var_list=var_list,
            outcome_name=outcome_name,
            cols_to_copy=cols_to_copy,
            params=params,
            imputation_map=imputation_map,
        )

    def merge_params(self, p: Optional[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Merge user parameters, returns new parameters does not alter object.

        :param p:
        :return: merged parameters
        """

        return vtreat_parameters(p)

    # noinspection PyPep8Naming
    def transform(self, X):
        """
        Apply transform to data.

        :param X: data
        :return: transformed data
        """

        X, orig_type = vtreat_impl.ready_data_frame(X)
        self.check_column_names(X.columns)
        if self.last_fit_x_id_ is None:
            raise ValueError("called transform on not yet fit treatment")
        if self.params_["check_for_duplicate_frames"] and (
            self.last_fit_x_id_ == vtreat.util.hash_data_frame(X)
        ):
            if self.params_["error_on_duplicate_frames"]:
                raise ValueError(
                    "possibly called transform on same data used to fit\n"
                    + "(this causes over-fit, please use fit_transform() instead)"
                )
            warnings.warn(
                "possibly called transform on same data used to fit\n"
                + "(this causes over-fit, please use fit_transform() instead)"
            )
        res = vtreat_impl.pre_prep_frame(
            X,
            col_list=self.plan_.num_list + self.plan_.cat_list,
            cols_to_copy=self.cols_to_copy_,
            cat_cols=self.plan_.cat_list,
        )
        res = vtreat_impl.perform_transform(x=res, transform=self, params=self.params_)
        res = vtreat_impl.limit_to_appropriate_columns(res=res, transform=self)
        res, res_columns = vtreat_impl.back_to_orig_type_data_frame(res, orig_type)
        self.last_result_columns = res_columns
        return res

    # noinspection PyPep8Naming
    def fit_transform(self, X, y=None, **fit_params):
        """
        fit_transform data, this is the way to fit with cross methods.

        :param X: explanatory values
        :param y: dependent values
        :param fit_params:
        :return: transformed data
        """

        X, orig_type = vtreat_impl.ready_data_frame(X)
        self.check_column_names(X.columns)
        if y is None:
            if self.outcome_name_ is None:
                raise ValueError(".fit_transform(X) must have outcome_name set")
            y = numpy.asarray(X[self.outcome_name_])
        else:
            y = numpy.asarray(y)
            if (self.outcome_name_ is not None) and (self.outcome_name_ in X.columns):
                if not numpy.all(X[self.outcome_name_] == y):
                    raise ValueError(
                        ".fit_transform(X, y) called with y != X[outcome_name]"
                    )
        if not X.shape[0] == len(y):
            raise ValueError("X.shape[0] should equal len(y)")
        y = vtreat.util.safe_to_numeric_array(y)
        if vtreat.util.is_bad(y).sum() > 0:
            raise ValueError("y should not have any missing/NA/NaN values")
        if numpy.max(y) <= numpy.min(y):
            raise ValueError("y does not vary")
        cross_rows = None
        cross_plan = None
        if self.params_["retain_cross_plan"]:
            cross_rows = self.cross_rows_
            cross_plan = self.cross_plan_
        self.clear()
        self.last_fit_x_id_ = vtreat.util.hash_data_frame(X)
        X = vtreat_impl.pre_prep_frame(
            X, col_list=self.var_list_, cols_to_copy=self.cols_to_copy_
        )
        if isinstance(y, pandas.Series):
            y = y.reset_index(inplace=False, drop=True)
        # model for independent transforms
        self.plan_ = None
        self.score_frame_ = None
        self.plan_ = vtreat_impl.fit_numeric_outcome_treatment(
            X=X,
            y=y,
            var_list=self.var_list_,
            outcome_name=self.outcome_name_,
            cols_to_copy=self.cols_to_copy_,
            params=self.params_,
            imputation_map=self.imputation_map_,
        )
        cross_frame = vtreat_impl.perform_transform(
            x=X, transform=self, params=self.params_
        )
        if (cross_plan is None) or (cross_rows != X.shape[0]):
            if cross_plan is not None:
                warnings.warn(
                    "Number of rows different than previous fit with retain_cross_plan==True"
                )
            cross_plan = self.params_["cross_validation_plan"].split_plan(
                n_rows=X.shape[0],
                k_folds=self.params_["cross_validation_k"],
                data=X,
                y=y,
            )
            cross_rows = X.shape[0]
        # patch in cross-frame versions of complex columns such as impact
        vtreat_impl.cross_patch_refit_y_aware_cols(
            x=X, y=y, res=cross_frame, plan=self.plan_, cross_plan=cross_plan
        )
        vtreat_impl.cross_patch_user_y_aware_cols(
            x=cross_frame,
            y=y,
            res=cross_frame,
            params=self.params_,
            cross_plan=cross_plan,
        )
        # use cross_frame to compute variable effects
        self.score_frame_ = vtreat_impl.score_plan_variables(
            cross_frame=cross_frame,
            outcome=y,
            plan=self.plan_,
            params=self.params_,
            is_classification=False,
        )
        if ("filter_to_recommended" in self.params_.keys()) and self.params_[
            "filter_to_recommended"
        ]:
            self.set_result_restriction(
                set(
                    [
                        ci
                        for ci in self.score_frame_["variable"][
                            self.score_frame_["recommended"]
                        ]
                    ]
                )
            )
        cross_frame = vtreat_impl.limit_to_appropriate_columns(
            res=cross_frame, transform=self
        )
        cross_frame, res_columns = vtreat_impl.back_to_orig_type_data_frame(
            cross_frame, orig_type
        )
        self.last_result_columns = res_columns
        if self.params_["retain_cross_plan"]:
            self.cross_plan_ = cross_plan
            self.cross_rows_ = cross_rows
        else:
            self.cross_plan_ = None
            self.cross_rows_ = None
        return cross_frame


class BinomialOutcomeTreatment(vtreat_impl.VariableTreatment):
    """manage a treatment plan for a target outcome (binomial classification)"""

    def __init__(
        self,
        *,
        var_list: Optional[Iterable[str]] = None,
        outcome_name: Optional[str] = None,
        outcome_target=None,
        cols_to_copy: Optional[Iterable[str]] = None,
        params: Optional[Dict[str, Any]] = None,
        imputation_map: Optional[Dict[str, Any]] = None,
    ):
        """

         :param var_list: list or tuple of column names, if empty all non outcome and copy columns are used
         :param outcome_name: name of column containing dependent variable
         :param outcome_target: value of outcome to consider "positive"
         :param cols_to_copy: list or tuple of column names
         :param params: vtreat.vtreat_parameters()
         :param imputation_map: map of column names to custom missing imputation values or functions
        """
        params = self.merge_params(params)
        vtreat_impl.VariableTreatment.__init__(
            self,
            var_list=var_list,
            outcome_name=outcome_name,
            outcome_target=outcome_target,
            cols_to_copy=cols_to_copy,
            params=params,
            imputation_map=imputation_map,
        )

    def merge_params(self, p: Optional[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Merge user parameters, returns new parameters does not alter object.

        :param p:
        :return: merged parameters
        """

        return vtreat_parameters(p)

    # noinspection PyPep8Naming
    def transform(self, X):
        """
        Apply transform to data.

        :param X: data
        :return: transformed data
        """

        X, orig_type = vtreat_impl.ready_data_frame(X)
        self.check_column_names(X.columns)
        if self.last_fit_x_id_ is None:
            raise ValueError("called transform on not yet fit treatment")
        if self.params_["check_for_duplicate_frames"] and (
            self.last_fit_x_id_ == vtreat.util.hash_data_frame(X)
        ):
            if self.params_["error_on_duplicate_frames"]:
                raise ValueError(
                    "possibly called transform on same data used to fit\n"
                    + "(this causes over-fit, please use fit_transform() instead)"
                )
            warnings.warn(
                "possibly called transform on same data used to fit\n"
                + "(this causes over-fit, please use fit_transform() instead)"
            )
        X = vtreat_impl.pre_prep_frame(
            X,
            col_list=self.plan_.num_list + self.plan_.cat_list,
            cols_to_copy=self.cols_to_copy_,
            cat_cols=self.plan_.cat_list,
        )
        res = vtreat_impl.perform_transform(x=X, transform=self, params=self.params_)
        res = vtreat_impl.limit_to_appropriate_columns(res=res, transform=self)
        res, res_columns = vtreat_impl.back_to_orig_type_data_frame(res, orig_type)
        self.last_result_columns = res_columns
        return res

    # noinspection PyPep8Naming
    def fit_transform(self, X, y=None, **fit_params):
        """
        fit_transform data, this is the way to fit with cross methods.

        :param X: explanatory values
        :param y: dependent values
        :param fit_params:
        :return: transformed data
        """

        X, orig_type = vtreat_impl.ready_data_frame(X)
        self.check_column_names(X.columns)
        if y is None:
            if self.outcome_name_ is None:
                raise ValueError(".fit_transform(X) must have outcome_name set")
            y = numpy.asarray(X[self.outcome_name_])
        else:
            y = numpy.asarray(y)
            if (self.outcome_name_ is not None) and (self.outcome_name_ in X.columns):
                if not numpy.all(X[self.outcome_name_] == y):
                    raise ValueError(
                        ".fit_transform(X, y) called with y != X[outcome_name]"
                    )
        if not X.shape[0] == len(y):
            raise ValueError("X.shape[0] should equal len(y)")
        y_mean = numpy.mean(y == self.outcome_target_)
        if y_mean <= 0 or y_mean >= 1:
            raise ValueError("y==outcome_target does not vary")
        cross_rows = None
        cross_plan = None
        if self.params_["retain_cross_plan"]:
            cross_rows = self.cross_rows_
            cross_plan = self.cross_plan_
        self.clear()
        self.last_fit_x_id_ = vtreat.util.hash_data_frame(X)
        X = vtreat_impl.pre_prep_frame(
            X, col_list=self.var_list_, cols_to_copy=self.cols_to_copy_
        )
        if isinstance(y, pandas.Series):
            y = y.reset_index(inplace=False, drop=True)
        # model for independent transforms
        self.plan_ = None
        self.score_frame_ = None
        self.plan_ = vtreat_impl.fit_binomial_outcome_treatment(
            X=X,
            y=y,
            outcome_target=self.outcome_target_,
            var_list=self.var_list_,
            outcome_name=self.outcome_name_,
            cols_to_copy=self.cols_to_copy_,
            params=self.params_,
            imputation_map=self.imputation_map_,
        )
        cross_frame = vtreat_impl.perform_transform(
            x=X, transform=self, params=self.params_
        )
        if (cross_plan is None) or (cross_rows != X.shape[0]):
            if cross_plan is not None:
                warnings.warn(
                    "Number of rows different than previous fit with retain_cross_plan==True"
                )
            cross_plan = self.params_["cross_validation_plan"].split_plan(
                n_rows=X.shape[0],
                k_folds=self.params_["cross_validation_k"],
                data=X,
                y=y,
            )
            cross_rows = X.shape[0]
        # patch in cross-frame versions of complex columns such as impact
        vtreat_impl.cross_patch_refit_y_aware_cols(
            x=X, y=y, res=cross_frame, plan=self.plan_, cross_plan=cross_plan
        )
        vtreat_impl.cross_patch_user_y_aware_cols(
            x=cross_frame,
            y=y,
            res=cross_frame,
            params=self.params_,
            cross_plan=cross_plan,
        )
        # use cross_frame to compute variable effects
        self.score_frame_ = vtreat_impl.score_plan_variables(
            cross_frame=cross_frame,
            outcome=numpy.asarray(
                numpy.asarray(y) == self.outcome_target_, dtype=float
            ),
            plan=self.plan_,
            params=self.params_,
            is_classification=True,
        )
        if ("filter_to_recommended" in self.params_.keys()) and self.params_[
            "filter_to_recommended"
        ]:
            self.set_result_restriction(
                set(
                    [
                        ci
                        for ci in self.score_frame_["variable"][
                            self.score_frame_["recommended"]
                        ]
                    ]
                )
            )
        cross_frame = vtreat_impl.limit_to_appropriate_columns(
            res=cross_frame, transform=self
        )
        cross_frame, res_columns = vtreat_impl.back_to_orig_type_data_frame(
            cross_frame, orig_type
        )
        self.last_result_columns = res_columns
        if self.params_["retain_cross_plan"]:
            self.cross_plan_ = cross_plan
            self.cross_rows_ = cross_rows
        else:
            self.cross_plan_ = None
            self.cross_rows_ = None
        return cross_frame


class MultinomialOutcomeTreatment(vtreat_impl.VariableTreatment):
    """manage a treatment plan for a set of outcomes (multinomial classification)"""

    def __init__(
        self,
        *,
        var_list: Optional[Iterable[str]] = None,
        outcome_name: Optional[str] = None,
        cols_to_copy: Optional[Iterable[str]] = None,
        params: Optional[Dict[str, Any]] = None,
        imputation_map: Optional[Dict[str, Any]] = None,
    ):
        """

         :param var_list: list or tuple of column names, if empty all non outcome and copy columns are used
         :param outcome_name: name of column containing dependent variable
         :param cols_to_copy: list or tuple of column names
         :param params: vtreat.vtreat_parameters()
         :param imputation_map: map of column names to custom missing imputation values or functions
        """

        params = self.merge_params(params)
        vtreat_impl.VariableTreatment.__init__(
            self,
            var_list=var_list,
            outcome_name=outcome_name,
            cols_to_copy=cols_to_copy,
            params=params,
            imputation_map=imputation_map,
        )
        self.outcomes_ = None

    def merge_params(self, p: Optional[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Merge user parameters, returns new parameters does not alter object.

        :param p:
        :return: merged parameters
        """

        return vtreat_parameters(p)

    # noinspection PyPep8Naming
    def transform(self, X):
        """
        Apply transform to data.

        :param X: data
        :return: transformed data
        """

        X, orig_type = vtreat_impl.ready_data_frame(X)
        self.check_column_names(X.columns)
        if self.last_fit_x_id_ is None:
            raise ValueError("called transform on not yet fit treatment")
        if self.params_["check_for_duplicate_frames"] and (
            self.last_fit_x_id_ == vtreat.util.hash_data_frame(X)
        ):
            if self.params_["error_on_duplicate_frames"]:
                raise ValueError(
                    "possibly called transform on same data used to fit\n"
                    + "(this causes over-fit, please use fit_transform() instead)"
                )
            warnings.warn(
                "possibly called transform on same data used to fit\n"
                + "(this causes over-fit, please use fit_transform() instead)"
            )
        X = vtreat_impl.pre_prep_frame(
            X,
            col_list=self.plan_.num_list + self.plan_.cat_list,
            cols_to_copy=self.cols_to_copy_,
            cat_cols=self.plan_.cat_list,
        )
        res = vtreat_impl.perform_transform(x=X, transform=self, params=self.params_)
        res = vtreat_impl.limit_to_appropriate_columns(res=res, transform=self)
        res, res_columns = vtreat_impl.back_to_orig_type_data_frame(res, orig_type)
        self.last_result_columns = res_columns
        return res

    # noinspection PyPep8Naming
    def fit_transform(self, X, y=None, **fit_params):
        """
        fit_transform data, this is the way to fit with cross methods.

        :param X: explanatory values
        :param y: dependent values
        :param fit_params:
        :return: transformed data
        """

        X, orig_type = vtreat_impl.ready_data_frame(X)
        self.check_column_names(X.columns)
        if y is None:
            if self.outcome_name_ is None:
                raise ValueError(".fit_transform(X) must have outcome_name set")
            y = numpy.asarray(X[self.outcome_name_])
        else:
            y = numpy.asarray(y)
            if (self.outcome_name_ is not None) and (self.outcome_name_ in X.columns):
                if not numpy.all(X[self.outcome_name_] == y):
                    raise ValueError(
                        ".fit_transform(X, y) called with y != X[outcome_name]"
                    )
        if not X.shape[0] == len(y):
            raise ValueError("X.shape[0] should equal len(y)")
        if len(numpy.unique(y)) <= 1:
            raise ValueError("y must take on at least 2 values")
        cross_rows = None
        cross_plan = None
        if self.params_["retain_cross_plan"]:
            cross_rows = self.cross_rows_
            cross_plan = self.cross_plan_
        self.clear()
        self.last_fit_x_id_ = vtreat.util.hash_data_frame(X)
        X = vtreat_impl.pre_prep_frame(
            X, col_list=self.var_list_, cols_to_copy=self.cols_to_copy_
        )
        if isinstance(y, pandas.Series):
            y = y.reset_index(inplace=False, drop=True)
        # model for independent transforms
        self.plan_ = None
        self.score_frame_ = None
        self.outcomes_ = numpy.unique(y)
        self.plan_ = vtreat_impl.fit_multinomial_outcome_treatment(
            X=X,
            y=y,
            var_list=self.var_list_,
            outcome_name=self.outcome_name_,
            cols_to_copy=self.cols_to_copy_,
            params=self.params_,
            imputation_map=self.imputation_map_,
        )
        cross_frame = vtreat_impl.perform_transform(
            x=X, transform=self, params=self.params_
        )
        if (cross_plan is None) or (cross_rows != X.shape[0]):
            if cross_plan is not None:
                warnings.warn(
                    "Number of rows different than previous fit with retain_cross_plan==True"
                )
            cross_plan = self.params_["cross_validation_plan"].split_plan(
                n_rows=X.shape[0],
                k_folds=self.params_["cross_validation_k"],
                data=X,
                y=y,
            )
            cross_rows = X.shape[0]
        vtreat_impl.cross_patch_refit_y_aware_cols(
            x=X, y=y, res=cross_frame, plan=self.plan_, cross_plan=cross_plan
        )
        vtreat_impl.cross_patch_user_y_aware_cols(
            x=cross_frame,
            y=y,
            res=cross_frame,
            params=self.params_,
            cross_plan=cross_plan,
        )
        # use cross_frame to compute variable effects

        def si(oi):
            """score i-th outcome group"""
            sf = vtreat_impl.score_plan_variables(
                cross_frame=cross_frame,
                outcome=numpy.asarray(numpy.asarray(y) == oi, dtype=float),
                plan=self.plan_,
                params=self.params_,
                is_classification=True,
            )
            sf["outcome_target"] = oi
            return sf

        score_frames = [si(oi) for oi in self.outcomes_]
        self.score_frame_ = pandas.concat(score_frames, axis=0)
        self.score_frame_.reset_index(inplace=True, drop=True)
        if ("filter_to_recommended" in self.params_.keys()) and self.params_[
            "filter_to_recommended"
        ]:
            self.set_result_restriction(
                set(
                    [
                        ci
                        for ci in self.score_frame_["variable"][
                            self.score_frame_["recommended"]
                        ]
                    ]
                )
            )
        cross_frame = vtreat_impl.limit_to_appropriate_columns(
            res=cross_frame, transform=self
        )
        cross_frame, res_columns = vtreat_impl.back_to_orig_type_data_frame(
            cross_frame, orig_type
        )
        self.last_result_columns = res_columns
        if self.params_["retain_cross_plan"]:
            self.cross_plan_ = cross_plan
            self.cross_rows_ = cross_rows
        else:
            self.cross_plan_ = None
            self.cross_rows_ = None
        return cross_frame


class UnsupervisedTreatment(vtreat_impl.VariableTreatment):
    """manage an unsupervised treatment plan"""

    def __init__(
        self,
        *,
        var_list: Optional[Iterable[str]] = None,
        cols_to_copy: Optional[Iterable[str]] = None,
        params: Optional[Dict[str, Any]] = None,
        imputation_map: Optional[Dict[str, Any]] = None,
    ):
        """

        :param var_list: list or tuple of column names, if empty all non copy columns are used
        :param cols_to_copy: list or tuple of column names
        :param params: vtreat.unsupervised_parameters()
        :param imputation_map: map of column names to custom missing imputation values or functions
        """
        params = self.merge_params(params)
        vtreat_impl.VariableTreatment.__init__(
            self,
            var_list=var_list,
            outcome_name=None,
            cols_to_copy=cols_to_copy,
            params=params,
            imputation_map=imputation_map,
        )

    def merge_params(self, p: Optional[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Merge user parameters, returns new parameters does not alter object.

        :param p:
        :return: merged parameters
        """

        return unsupervised_parameters(p)

    # noinspection PyPep8Naming
    def transform(self, X):
        """
        Apply transform to data.

        :param X: data
        :return: transformed data
        """

        X, orig_type = vtreat_impl.ready_data_frame(X)
        self.check_column_names(X.columns)
        if self.last_fit_x_id_ is None:
            raise ValueError("called transform on not yet fit treatment")
        X = vtreat_impl.pre_prep_frame(
            X,
            col_list=self.plan_.num_list + self.plan_.cat_list,
            cols_to_copy=self.cols_to_copy_,
            cat_cols=self.plan_.cat_list,
        )
        res = vtreat_impl.perform_transform(x=X, transform=self, params=self.params_)
        res = vtreat_impl.limit_to_appropriate_columns(res=res, transform=self)
        res, res_columns = vtreat_impl.back_to_orig_type_data_frame(res, orig_type)
        self.last_result_columns = res_columns
        return res

    # noinspection PyPep8Naming
    def fit_transform(self, X, y=None, **fit_params):
        """
        fit_transform data.

        :param X: explanatory values
        :param y: dependent values
        :param fit_params:
        :return: transformed data
        """

        X, orig_type = vtreat_impl.ready_data_frame(X)
        self.check_column_names(X.columns)
        assert y is None
        self.clear()
        self.last_fit_x_id_ = vtreat.util.hash_data_frame(X)
        X = vtreat_impl.pre_prep_frame(
            X, col_list=self.var_list_, cols_to_copy=self.cols_to_copy_
        )
        self.plan_ = vtreat_impl.fit_unsupervised_treatment(
            X=X,
            var_list=self.var_list_,
            outcome_name=self.outcome_name_,
            cols_to_copy=self.cols_to_copy_,
            params=self.params_,
            imputation_map=self.imputation_map_,
        )
        res = vtreat_impl.perform_transform(x=X, transform=self, params=self.params_)
        self.score_frame_ = vtreat_impl.pseudo_score_plan_variables(
            cross_frame=res, plan=self.plan_, params=self.params_
        )
        if ("filter_to_recommended" in self.params_.keys()) and self.params_[
            "filter_to_recommended"
        ]:
            self.set_result_restriction(
                set(
                    [
                        ci
                        for ci in self.score_frame_["variable"][
                            self.score_frame_["recommended"]
                        ]
                    ]
                )
            )
        res = vtreat_impl.limit_to_appropriate_columns(res=res, transform=self)
        res, res_columns = vtreat_impl.back_to_orig_type_data_frame(res, orig_type)
        self.last_result_columns = res_columns
        return res



"""base class for user transforms"""

import abc


class UserTransform(abc.ABC):
    """base class for user transforms, should express taking a set of k inputs to k outputs independently"""

    def __init__(self, treatment):
        self.y_aware_ = True
        self.treatment_ = treatment
        self.incoming_vars_ = []
        self.derived_vars_ = []

    # noinspection PyPep8Naming
    @abc.abstractmethod
    def fit(self, X, y):
        """
        sklearn API

        :param X: explanatory values
        :param y: dependent values
        :return: self for method chaining
        """

    # noinspection PyPep8Naming
    @abc.abstractmethod
    def transform(self, X):
        """
        sklearn API

        :param X: explanatory values
        :return: transformed data
        """

    # noinspection PyPep8Naming
    def fit_transform(self, X, y):
        """
        sklearn API

        :param X: explanatory values
        :param y: dependent values
        :return: transformed data
        """

        self.fit(X, y)
        return self.transform(X)



"""
vtreat data preparation package: https://github.com/WinVector/pyvtreat
"""

# noinspection PyUnresolvedReferences
import pandas

# noinspection PyUnresolvedReferences
import numpy

from vtreat.vtreat_api import *

__docformat__ = "restructuredtext"
__version__ = "1.2.4"

__doc__ = """
This<https://github.com/WinVector/pyvtreat> is the Python version of the vtreat data preparation system
(also available as an R package<https://winvector.github.io/vtreat/>.

vtreat is a DataFrame processor/conditioner that prepares
real-world data for supervised machine learning or predictive modeling
in a statistically sound manner.

vtreat takes an input DataFrame
that has a specified column called "the outcome variable" (or "y")
that is the quantity to be predicted (and must not have missing
values).  Other input columns are possible explanatory variables
(typically numeric or categorical/string-valued, these columns may
have missing values) that the user later wants to use to predict "y".
In practice such an input DataFrame may not be immediately suitable
for machine learning procedures that often expect only numeric
explanatory variables, and may not tolerate missing values.

To solve this, vtreat builds a transformed DataFrame where all
explanatory variable columns have been transformed into a number of
numeric explanatory variable columns, without missing values.  The
vtreat implementation produces derived numeric columns that capture
most of the information relating the explanatory columns to the
specified "y" or dependent/outcome column through a number of numeric
transforms (indicator variables, impact codes, prevalence codes, and
more).  This transformed DataFrame is suitable for a wide range of
supervised learning methods from linear regression, through gradient
boosted machines.

The idea is: you can take a DataFrame of messy real world data and
easily, faithfully, reliably, and repeatably prepare it for machine
learning using documented methods using vtreat.  Incorporating
vtreat into your machine learning workflow lets you quickly work
with very diverse structured data.

Worked examples can be found `here`<https://github.com/WinVector/pyvtreat/tree/master/Examples>.

For more detail please see here: `arXiv:1611.09477
stat.AP`<https://arxiv.org/abs/1611.09477> (the documentation describes the R version,
however all of the examples can be found worked in Python 
`here`<https://github.com/WinVector/pyvtreat/tree/master/Examples/vtreat_paper1>).

vtreat is available
as a `Python/Pandas package`<https://github.com/WinVector/vtreat>,
and also as an `R package`<https://github.com/WinVector/vtreat>.
"""



"""Basic cross validation methods"""

import abc
from typing import Dict, List, Optional
import warnings

import numpy
import numpy.random

import pandas


class CrossValidationPlan(abc.ABC):
    """Data splitting plan"""

    def __init__(self):
        pass

    @abc.abstractmethod
    def split_plan(
        self,
        *,
        n_rows: Optional[int] = None,
        k_folds: Optional[int] = None,
        data=None,
        y=None,
    ) -> List[Dict[str, List[int]]]:
        """
        Build a cross validation plan for the given parameters.

        :param n_rows: (optional) number of input rows
        :param k_folds: (optional) number of folds we want
        :param data: (optional) explanatory variables
        :param y: (optional) dependent variable
        :return: cross validation plan (list of dictionaries)
        """


def _k_way_cross_plan_y_stratified(
    *, n_rows: int, k_folds: int, y
) -> List[Dict[str, List[int]]]:
    """

    :param n_rows: number of input rows
    :param k_folds: number of cross folds desired
    :param y: values to stratify on
    :return: cross validation plan (list of dictionaries)
    """
    """randomly split range(n_rows) into k_folds disjoint groups, attempting an even y-distribution"""
    if k_folds < 2:
        k_folds = 2
    if k_folds > n_rows:
        k_folds = n_rows
    if n_rows <= 2 or k_folds <= 1:
        # degenerate overlap cases
        warnings.warn("degenerate cross plan, not disjoint", UserWarning)
        plan = [
            {"train": [i for i in range(n_rows)], "app": [i for i in range(n_rows)]}
        ]
        return plan
    # first sort by y plus a random key
    if y is None:
        y = numpy.zeros(n_rows)
    else:
        y = numpy.asarray(y)
    assert len(y) == n_rows
    d = pandas.DataFrame(
        {
            "y": y,
            "i": [i for i in range(n_rows)],
            "r": numpy.random.uniform(size=n_rows),
        }
    )
    d.sort_values(by=["y", "r"], inplace=True)
    d.reset_index(inplace=True, drop=True)
    # assign y-blocks to lose fine details of y
    fold_size = n_rows / k_folds
    d["block"] = [numpy.floor(i / fold_size) for i in range(n_rows)]
    d.sort_values(by=["block", "r"], inplace=True)
    d.reset_index(inplace=True, drop=True)
    # now assign groups modulo k (ensuring at least one in each group)
    d["grp"] = [i % k_folds for i in range(n_rows)]
    d.sort_values(by=["i"], inplace=True)
    d.reset_index(inplace=True, drop=True)
    grp = numpy.asarray(d["grp"])
    plan = [
        {
            "train": [i for i in range(n_rows) if grp[i] != j],
            "app": [i for i in range(n_rows) if grp[i] == j],
        }
        for j in range(k_folds)
    ]
    return plan


class KWayCrossPlanYStratified(CrossValidationPlan):
    """K-way cross validation plan, attempting an even y-distribution"""

    def __init__(self):
        CrossValidationPlan.__init__(self)

    def split_plan(
        self,
        *,
        n_rows: Optional[int] = None,
        k_folds: Optional[int] = None,
        data=None,
        y=None,
    ) -> List[Dict[str, List[int]]]:
        """

        :param n_rows: required, number of rows
        :param k_folds: required, number of cross-folds
        :param data: not used
        :param y: required, outcomes to stratify on
        :return:
        """
        if n_rows is None:
            raise ValueError("n_rows must not be None")
        if k_folds is None:
            raise ValueError("k_folds must not be None")
        if y is None:
            y = numpy.zeros(n_rows)
        assert len(y) == n_rows
        return _k_way_cross_plan_y_stratified(n_rows=n_rows, k_folds=k_folds, y=y)



"""
Utility functions for vtreat
"""

import math
import re
import statistics
from typing import Iterable, List, Optional

import hashlib

import numpy
import pandas

import vtreat.stats_utils


def safe_to_numeric_array(x):
    """
    Convert array to numeric. Note, will parse strings (due to numpy)!

    :param x: array to process
    :return: numeric array
    """
    # note will parse strings!
    return numpy.asarray(x, dtype=float)


def can_convert_v_to_numeric(x) -> bool:
    """
    check if non-empty vector can convert to numeric

    :param x:
    :return: True if can convert to numeric, false otherwise (no string parsing).
    """
    x = numpy.asarray(x)
    not_bad = numpy.logical_not(pandas.isnull(x))
    n_not_bad = numpy.sum(not_bad)
    if n_not_bad < 1:
        return True   # All nan/None, can convert to numeric
    try:
        numpy.asarray(
            x[not_bad] + 0, dtype=float
        )  # +0 prevents string parse, leaving strings as strings
        return True
    except TypeError:
        return False


def is_bad(x):
    """
    For numeric vector x, return logical vector of positions that are null, NaN, infinite.

    :param x:
    :return:
    """
    if can_convert_v_to_numeric(x):
        x = safe_to_numeric_array(x)
        return numpy.logical_or(
            pandas.isnull(x), numpy.logical_or(numpy.isnan(x), numpy.isinf(x))
        )
    return pandas.isnull(x)


def numeric_has_range(x) -> bool:
    """
    Check if a numeric vector has numeric range.

    :param x: vector to check
    :return: True if max > min values in vector, else False.
    """
    x = safe_to_numeric_array(x)
    not_bad = numpy.logical_not(is_bad(x))
    n_not_bad = numpy.sum(not_bad)
    if n_not_bad < 2:
        return False
    x = x[not_bad]
    return numpy.max(x) > numpy.min(x)


def summarize_column(x, *, fn=numpy.mean) -> float:
    """
    Summarize column to a non-missing scalar.

    :param x: a vector/Series or column of numbers
    :param fn: summarize function (such as numpy.mean), only passed non-bad positions
    :return: scalar float summary of the non-None positions of x (otherwise 0)
    """

    x = safe_to_numeric_array(x)
    not_bad = numpy.logical_not(is_bad(x))
    n_not_bad = numpy.sum(not_bad)
    if n_not_bad < 1:
        return 0.0
    x = x[not_bad]
    v = 0.0 + fn(x)
    if pandas.isnull(v) or math.isnan(v) or math.isinf(v):
        return 0.0
    return v


def get_unique_value_count(x):
    """compute how many unique values in list-x"""
    if len(x) <= 1:
        return len(x)
    p = pandas.DataFrame({"x": x, "o": 1})
    s = p.groupby("x").sum()  # drops None
    return numpy.max(pandas.isnull(x)) + s.shape[0]


def grouped_by_x_statistics(x, y):
    """compute some grouped by x vector summaries of numeric y vector (no missing values in y)"""
    n = len(x)
    assert n > 0
    assert n == len(y)
    y = safe_to_numeric_array(y)
    eps = 1.0e-3
    sf = pandas.DataFrame({"x": x, "y": y})
    sf.reset_index(inplace=True, drop=True)
    bad_posns = pandas.isnull(sf["x"])
    sf.loc[bad_posns, "x"] = "_NA_"
    global_mean = sf["y"].mean()
    sf["_group_mean"] = sf.groupby("x")["y"].transform("mean")
    sf["_var"] = (sf["y"] - sf["_group_mean"]) ** 2
    sf["_ni"] = 1
    sf = sf.groupby("x").sum()
    sf.reset_index(inplace=True, drop=False)
    sf["y"] = sf["y"] / sf["_ni"]
    sf["_group_mean"] = sf["_group_mean"] / sf["_ni"]
    sf["_var"] = sf["_var"] / (sf["_ni"] - 1) + eps
    avg_var = 0
    bad_vars = is_bad(sf["_var"])
    if numpy.sum(bad_vars) < len(sf["_var"]):
        avg_var = numpy.nanmean(sf["_var"])
    sf.loc[bad_vars, "_var"] = avg_var
    if sf.shape[0] > 1:
        sf["_vb"] = statistics.variance(sf["_group_mean"]) + eps
    else:
        sf["_vb"] = eps
    sf["_gm"] = global_mean
    # hierarchical model is in:
    # http://www.win-vector.com/blog/2017/09/partial-pooling-for-lower-variance-variable-encoding/
    # using naive empirical estimates of variances
    # adjusted from ni to ni-1 and +eps variance to make
    # rare levels look like new levels.
    sf["_hest"] = (
        (sf["_ni"] - 1) * sf["_group_mean"] / sf["_var"] + sf["_gm"] / sf["_vb"]
    ) / ((sf["_ni"] - 1) / sf["_var"] + 1 / sf["_vb"])
    return sf


def score_variables(
        cross_frame: pandas.DataFrame,
        *,
        variables: Optional[Iterable[str]],
        outcome,
        is_classification: bool = False) -> Optional[pandas.DataFrame]:
    """
    Score the linear relation of variables to outcome.

    :param cross_frame: data frame to score
    :param variables: names of variables to score
    :param outcome: numeric outcome
    :param is_classification: logical, True if a classification problem (0/1)
    :return: data frame of variable evaluations, None if there are no non-constant variables.
    """

    assert isinstance(cross_frame, pandas.DataFrame)
    if variables is not None:
        assert not isinstance(variables, str)  # prevents str masquerading as iterable
        variables = list(variables)
    else:
        variables = list(cross_frame.columns)
    if len(variables) <= 0:
        return None
    outcome = safe_to_numeric_array(outcome)
    n = cross_frame.shape[0]
    assert n == len(outcome)
    if numpy.max(outcome) <= numpy.min(outcome):
        return None  # y must have range

    def f(v):
        """summarizing fn"""
        col = cross_frame[v]
        if (
            (n > 2)
            and (numpy.max(col) > numpy.min(col))
            and (numpy.max(outcome) > numpy.min(outcome))
        ):
            cor, sig = vtreat.stats_utils.our_corr_score(y_true=outcome, y_pred=col)
            r2 = cor ** 2
            if is_classification:
                r2, sig = vtreat.stats_utils.our_pseudo_R2(y_true=outcome, y_pred=col)
            sfi = pandas.DataFrame(
                {
                    "variable": [v],
                    "has_range": [True],
                    "PearsonR": [cor],
                    "R2": [r2],
                    "significance": [sig],
                }
            )
        else:
            sfi = pandas.DataFrame(
                {
                    "variable": [v],
                    "has_range": [False],
                    "PearsonR": [numpy.NaN],
                    "R2": [numpy.NaN],
                    "significance": [1.0],
                }
            )
        return sfi

    sf_list = [f(v) for v in variables]
    sf = pandas.concat(sf_list, axis=0, sort=False)
    sf = sf.reset_index(inplace=False, drop=True)
    return sf


def check_matching_numeric_frames(*, res, expect, tol=1.0e-4):
    """
    Check if two numeric pandas.DataFrame s are identical.  assert if not
    :param res:
    :param expect:
    :param tol: numeric tolerance.
    :return: None
    """
    assert isinstance(expect, pandas.DataFrame)
    assert isinstance(res, pandas.DataFrame)
    assert res.shape == expect.shape
    for c in expect.columns:
        ec = expect[c]
        rc = res[c]
        assert numpy.max(numpy.abs(ec - rc)) <= tol


def unique_items_in_order(x) -> list:
    """
    Return de-duplicated list of items in order they are in supplied array.

    :param x: vector to inspect
    :return: list
    """
    ret = []
    if x is not None:
        seen = set()
        for item in x:
            if item not in seen:
                ret.append(item)
                seen.add(item)
    return ret


def clean_string(s: str) -> str:
    """
    Replace common symbols with column-name safe alternatives.

    :param s: incoming string
    :return: string
    """
    mp = {
        "<": "_lt_",
        ">": "_gt_",
        "[": "_osq_",
        "]": "_csq_",
        "(": "_op_",
        ")": "_cp_",
        ".": "_",
        " ": "_",
    }
    s = s.strip()
    s = re.sub(r"\s+", " ", s)
    for (k, v) in mp.items():
        s = s.replace(k, v)
    return s


def build_level_codes(incoming_column_name: str, levels: Iterable) -> List[str]:
    """
    Pick level names for a set of levels.

    :param incoming_column_name:
    :param levels:
    :return:
    """
    levels = [str(lev) for lev in levels]
    levels = [incoming_column_name + "_lev_" + clean_string(lev) for lev in levels]
    if len(set(levels)) != len(levels):
        levels = [levels[i] + "_" + str(i) for i in range(len(levels))]
    return levels


def hash_data_frame(d: pandas.DataFrame) -> str:
    """
    Get a hash code representing a data frame.

    :param d: data frame
    :return: hash code as a string
    """
    return hashlib.sha256(pandas.util.hash_pandas_object(d).values).hexdigest()






"""
Convert the description of a vtreat variable treatment into a data algebra pipeline.
"""

from typing import Dict, Iterable, List, Optional, Tuple
import numpy
import pandas

from vtreat.vtreat_impl import bad_sentinel, replace_bad_with_sentinel


from data_algebra.data_ops import data, descr, describe_table, TableDescription, ViewRepresentation
from data_algebra.solutions import def_multi_column_map


def _check_treatment_table(vtreat_descr: pandas.DataFrame):
    """
    Assert if expected invariants don't hold for vtreat_descr.

    :param vtreat_descr: .description_matrix() description of a transform to check.
    :return: no return, assert on failure
    """

    # belt and suspenders replace missing with sentinel
    vtreat_descr = vtreat_descr.copy()
    vtreat_descr["value"] = replace_bad_with_sentinel(vtreat_descr["value"])
    # check our expected invariants
    assert isinstance(vtreat_descr, pandas.DataFrame)
    # numeric is a function of original variable only
    check_fn_relnn = (
        data(vtreat_descr=vtreat_descr)
            .project({}, group_by=["orig_var", "orig_was_numeric"])
            .extend({"one": 1})
            .project({"count": "one.sum()"}, group_by=["orig_var"])
    ).ex()
    assert numpy.all(check_fn_relnn["count"] == 1)
    # variable consumed is function of variable produced and treatment only
    check_fn_reln2 = (
        data(vtreat_descr=vtreat_descr)
            .project({}, group_by=["treatment", "orig_var", "variable"])
            .extend({"one": 1})
            .project({"count": "one.sum()"}, group_by=["treatment", "variable"])
    ).ex()
    assert numpy.all(check_fn_reln2["count"] == 1)
    # clean copies don't change variable names
    cn_rows = vtreat_descr.loc[
        vtreat_descr["treatment_class"] == "CleanNumericTransform", :
    ].reset_index(inplace=False, drop=True)
    assert numpy.all(cn_rows["variable"] == cn_rows["orig_var"])
    # operations other than clean copy produce new variable names
    ot_rows = vtreat_descr.loc[
        vtreat_descr["treatment_class"] != "CleanNumericTransform", :
    ].reset_index(inplace=False, drop=True)
    assert len(set(ot_rows["variable"]).intersection(vtreat_descr["orig_var"])) == 0
    # clean copy and re-mapping take disjoint inputs (one alters input as a prep-step, so they would interfere)
    mp_rows = (
        data(vtreat_descr=vtreat_descr)
        .select_rows("treatment_class == 'MappedCodeTransform'")
        .project({}, group_by=["orig_var", "variable"])
        .order_rows(["orig_var", "variable"])
    ).ex()
    assert len(set(mp_rows["orig_var"]).intersection(cn_rows["orig_var"])) == 0


def as_data_algebra_pipeline(
    *,
    source: TableDescription,
    vtreat_descr: pandas.DataFrame,
    treatment_table_name: str,
    row_keys: Iterable[str],
) -> ViewRepresentation:
    """
    Convert the description of a vtreat transform (gotten via .description_matrix())
    into a data algebra pipeline.
    See: https://github.com/WinVector/data_algebra and https://github.com/WinVector/pyvtreat .
    Missing and nan are treated as synonyms for '_NA_'.
    Assembling the entire pipeline can be expensive. If one is willing to instantiate tables
    it can be better to sequence operations instead of composing them.
    Another way to use this methodology would be to port this code as a stored procedure
    in a target database of choice, meaning only the vtreat_descr table would be needed on such systems.

    :param source: input data.
    :param vtreat_descr: .description_matrix() description of transform.
                         Expected invariant: CleanNumericTransform doesn't change variable names,
                         all other operations produce new names.
    :param treatment_table_name: name to use for the vtreat_descr table.
    :param row_keys: list of columns uniquely keying rows
    :return: data algebra pipeline implementing specified vtreat treatment
    """

    assert isinstance(source, TableDescription)
    assert isinstance(vtreat_descr, pandas.DataFrame)
    assert isinstance(treatment_table_name, str)
    assert row_keys is not None
    assert not isinstance(row_keys, str)
    row_keys = list(row_keys)
    assert len(row_keys) > 0
    assert numpy.all([isinstance(v, str) for v in row_keys])

    _check_treatment_table(vtreat_descr)
    # belt and suspenders replace missing with sentinel
    vtreat_descr = vtreat_descr.copy()
    vtreat_descr["value"] = replace_bad_with_sentinel(vtreat_descr["value"])
    # start building up operator pipeline
    ops = source
    step_1_ops = dict()
    # add in is_bad indicators
    im_rows = vtreat_descr.loc[
        vtreat_descr["treatment_class"] == "IndicateMissingTransform", :
    ].reset_index(inplace=False, drop=True)
    for i in range(im_rows.shape[0]):
        if im_rows['orig_was_numeric'][i]:
            step_1_ops[
                im_rows["variable"][i]
            ] = f"{im_rows['orig_var'][i]}.is_bad().where(1.0, 0.0)"
        else:
            step_1_ops[
                im_rows["variable"][i]
            ] = f"({im_rows['orig_var'][i]}.coalesce('{bad_sentinel}') == '{bad_sentinel}').where(1.0, 0.0)"
    # add in general value indicators or dummies, all indicators are non-numeric (string)
    ic_rows = vtreat_descr.loc[
        vtreat_descr["treatment_class"] == "IndicatorCodeTransform", :
    ].reset_index(inplace=False, drop=True)
    for i in range(ic_rows.shape[0]):
        ov = ic_rows["orig_var"].values[i]
        vi = ic_rows["value"].values[i]
        step_1_ops[
            ic_rows["variable"][i]
        ] = f"({ov}.coalesce('{bad_sentinel}') == '{vi}').where(1.0, 0.0)"
    if len(step_1_ops) > 0:
        ops = ops.extend(step_1_ops)
    # mapped columns
    mapping_table = (
        describe_table(vtreat_descr, table_name=treatment_table_name)
            .select_rows('treatment_class == "MappedCodeTransform"')
            .select_columns(['orig_var', 'value', 'replacement', 'treatment']))
    mapping_rows = mapping_table.transform(vtreat_descr)
    if mapping_rows.shape[0] > 0:
        groups = list(set(mapping_rows['treatment']))
        mapping_rows = mapping_rows.groupby('treatment')
        for group_name in groups:
            mg = mapping_rows.get_group(group_name)
            if mg.shape[0] > 0:
                cols_to_map = list(set(mg['orig_var']))
                cols_to_map_back = [f'{c}_{group_name}' for c in cols_to_map]
                ops_g = def_multi_column_map(
                    source.extend({v: f"{v}.coalesce('{bad_sentinel}')" for v in cols_to_map}),
                    mapping_table=mapping_table.select_rows(f'treatment == "{group_name}"'),
                    row_keys=row_keys,
                    cols_to_map=cols_to_map,
                    cols_to_map_back=cols_to_map_back,
                    coalesce_value=0.0,
                    col_name_key='orig_var',
                    col_value_key='value',
                    mapped_value_key='replacement',
                )
                ops = ops.natural_join(
                    b=ops_g,
                    by=row_keys,
                    jointype='left',
                )
    # add in any clean numeric copies, inputs are numeric- so disjoint of categorical processing
    cn_rows = vtreat_descr.loc[
        vtreat_descr["treatment_class"] == "CleanNumericTransform", :
    ].reset_index(inplace=False, drop=True)
    if cn_rows.shape[0] > 0:
        step_3_exprs = dict()
        for i in range(cn_rows.shape[0]):
            step_3_exprs[
                cn_rows["variable"][i]
            ] = f"{cn_rows['orig_var'][i]}.coalesce({cn_rows['replacement'][i]})"
        ops = ops.extend(step_3_exprs)
    # remove any input variables that are not the same name as variables we produced
    # this prevents non-numerics from leaking forward
    to_del = list(set(vtreat_descr["orig_var"]) - set(vtreat_descr["variable"]))
    if len(to_del) > 0:
        to_del.sort()
        ops = ops.drop_columns(to_del)
    return ops



"""util for basic statistical steps"""

from typing import Tuple

import numpy
import pandas

import scipy.stats
import scipy.optimize
import sklearn.linear_model


# methods to avoid calling statsmodels which seems to be incompatible with many
# versions of other packages we need:
#  https://github.com/WinVector/pyvtreat/issues/14


def our_corr_score(*, y_true, y_pred) -> Tuple[float, float]:
    """
    Compute Pearson correlation. Case-out some corner cases.

    :param y_true: truth values
    :param y_pred: predictions
    :return: (pearson r, significance)
    """

    if not isinstance(y_true, numpy.ndarray):
        y_true = numpy.asarray(y_true)
    if not isinstance(y_pred, numpy.ndarray):
        y_pred = numpy.asarray(y_pred)
    n = len(y_true)
    if n < 2:
        return 1, 1
    if numpy.min(y_true) >= numpy.max(y_true):
        return 1, 1
    if numpy.min(y_pred) >= numpy.max(y_pred):
        return 0, 1
    r, sig = scipy.stats.pearsonr(y_true, y_pred)
    if n < 3:
        sig = 1
    return r, sig


def est_deviance(*, y, est, epsilon: float = 1.0e-5) -> float:
    """
    Estimate the deviance

    :param y: truth values
    :param est: predictions
    :param epsilon: how close to get to 0 and 1
    :return: deviance estimate
    """

    if not isinstance(y, numpy.ndarray):
        y = numpy.asarray(y)
    if not isinstance(est, numpy.ndarray):
        est = numpy.asarray(est)
    est = numpy.minimum(est, 1 - epsilon)
    est = numpy.maximum(est, epsilon)
    deviance = -2 * numpy.sum(y * numpy.log(est) + (1 - y) * numpy.log(1 - est))
    return deviance


def sklearn_solve_logistic(*, y, x, regularization: float = 1.0e-6):
    """
    Single variable logistic regression.
    Assumes special cases of solve_logistic_regression already eliminated.

    :param y: dependent variable
    :param x: explanatory variable
    :param regularization:
    :return: model predictions
    """

    if not isinstance(y, numpy.ndarray):
        y = numpy.asarray(y)
    if not isinstance(x, numpy.ndarray):
        x = numpy.asarray(x)
    fitter = sklearn.linear_model.LogisticRegression(
        penalty="l2", solver="lbfgs", fit_intercept=True, C=1 / regularization
    )
    dependent_vars = x.reshape((len(y), 1))
    fitter.fit(X=dependent_vars, y=y)
    preds = fitter.predict_proba(X=dependent_vars)[:, 1]
    return preds


# x, y - numpy numeric vectors, y 0/1.  solve for y- return predictions
def solve_logistic_regression(*, y, x):
    """
    Single variable logistic regression. Returns predictions, corner
    cases removed.

    :param y: dependent variable
    :param x: explanatory variable
    :return: predictions
    """

    if not isinstance(y, numpy.ndarray):
        y = numpy.asarray(y)
    if not isinstance(x, numpy.ndarray):
        x = numpy.asarray(x)
    # catch some corner cases
    n = len(y)
    if (n < 2) or (numpy.min(y) >= numpy.max(y)):
        return y.copy()
    if numpy.min(x) >= numpy.max(x):
        return numpy.asarray([numpy.mean(y)] * n)
    # check for fully separable cases
    big_y_indices = y > 0
    x_b = x[big_y_indices]
    x_s = x[numpy.logical_not(big_y_indices)]
    if (numpy.min(x_b) > numpy.max(x_s)) or (numpy.max(x_b) < numpy.min(x_s)):
        r = numpy.zeros(n)
        r[big_y_indices] = 1
        return r
    # run a full logistic regression
    preds = sklearn_solve_logistic(y=y, x=x)
    return numpy.asarray(preds)


# noinspection PyPep8Naming
def our_pseudo_R2(*, y_true, y_pred) -> Tuple[float, float]:
    """
    Return the logistic pseudo-R2

    :param y_true: dependent variable
    :param y_pred: explanatory variable
    :return: (pseudo-R2, significance)
    """

    if not isinstance(y_true, numpy.ndarray):
        y_true = numpy.asarray(y_true)
    if not isinstance(y_pred, numpy.ndarray):
        y_pred = numpy.asarray(y_pred)
    n = len(y_true)
    if n < 2:
        return 1, 1
    if numpy.min(y_true) >= numpy.max(y_true):
        return 1, 1
    if numpy.min(y_pred) >= numpy.max(y_pred):
        return 0, 1
    preds = solve_logistic_regression(y=y_true, x=y_pred)
    deviance = est_deviance(y=y_true, est=preds)
    null_deviance = est_deviance(y=y_true, est=numpy.zeros(n) + numpy.mean(y_true))
    r2 = 1 - deviance / null_deviance
    sig = 1
    if n >= 3:
        # https://github.com/WinVector/sigr/blob/master/R/ChiSqTest.R
        df_null = n - 1
        df_residual = n - 2
        delta_deviance = null_deviance - deviance
        delta_df = df_null - df_residual
        sig = 1 - scipy.stats.chi2.cdf(x=delta_deviance, df=delta_df)
    return r2, sig


def xicor(xvec, yvec, *, n_reps: int = 5) -> Tuple[float, float]:
    """
    xicor calculation built to match from R::caclulateXI() and published article.

    :param xvec: numeric vector with explanatory variable to compute xicor for.
    :param yvec: numeric dependent variable to relate to.
    :param n_reps: number of times to repeat calculation.
    :return: mean and standard error of estimate (under x-tie breaking)
    """
    yvec = numpy.asarray(yvec)
    n = len(yvec)
    assert n > 1
    assert isinstance(n_reps, int)
    assert n_reps > 0
    xvec = numpy.asarray(xvec)
    assert n == len(xvec)
    PI = numpy.zeros(n)
    fr_orig = scipy.stats.rankdata(yvec, method="max") / n
    gr = scipy.stats.rankdata(-yvec, method="max") / n
    CU = numpy.mean(gr * (1 - gr))
    xi_s = numpy.zeros(n_reps)
    for rep_i in range(n_reps):
        perm = numpy.random.permutation(n)  # get random ranking by ranking permuted
        PI_inv = scipy.stats.rankdata(xvec[perm], method="ordinal")
        PI[perm] = PI_inv  # invert permutation, self assignment fails
        ord = numpy.argsort(PI)
        fr = fr_orig[ord]
        A1 = numpy.sum(numpy.abs(fr[0:(n - 1)] - fr[1:n])) / (2 * n)
        xi = 1 - A1 / CU
        xi_s[rep_i] = xi
    return numpy.mean(xi_s), numpy.std(xi_s) / numpy.sqrt(n_reps)


def xicor_for_frame(d: pandas.DataFrame, y, *, n_reps=5):
    """
    Calculate xicor for all columns of data frame d with respect to dependent column y.

    :param d: data frame of proposed explanatory variables.
    :param y: vector of the dependent variable values.
    :param n_reps: number of times to repeat experiment (positive integer)
    :return: data frame with: variable (name of column), xicor (estimated xicor statistic),
             xicor_se (standard error of xicor estimate, goes to zero as n_reps grows),
             xicor_perm_mean (mean value of xicor with y scrambled, goes to zero as n_reps grows),
             xicor_perm_stddev (sample standard deviation of y scrambled xicor estimates,
             used to form z or t style estimates).
    """
    n = d.shape[0]
    assert n > 1
    y = numpy.asarray(y)
    assert len(y) == n
    assert isinstance(n_reps, int)
    assert n_reps > 0
    res = pandas.DataFrame({
        'variable': d.columns,
        'xicor': 0.0,
        'xicor_se': 0.0,
        'xicor_perm_mean': 0.0,
        'xicor_perm_stddev': 0.0,
        'xicor_perm_sum': 0.0,
        'xicor_perm_sum_sq': 0.0,
    })
    # get the xicor estimates
    for col_i in range(len(d.columns)):
        xvec = d[d.columns[col_i]]
        xi_est, xi_est_dev = xicor(xvec, y, n_reps=n_reps)
        res.loc[col_i, 'xicor'] = xi_est
        res.loc[col_i, 'xicor_se'] = xi_est_dev
    # score all x-columns with the same y-permutation
    # estimate stddev with expanding squares to cut down storage
    for rep_j in range(n_reps):
        y_perm = y[numpy.random.permutation(n)]
        for col_i in range(len(d.columns)):
            xvec = d[d.columns[col_i]]
            xi_perm, _ = xicor(xvec, y_perm, n_reps=1)
            res.loc[col_i, 'xicor_perm_sum'] = res.loc[col_i, 'xicor_perm_sum'] + xi_perm
            res.loc[col_i, 'xicor_perm_sum_sq'] = res.loc[col_i, 'xicor_perm_sum_sq'] + xi_perm * xi_perm
    res['xicor_perm_mean'] = res['xicor_perm_sum'] / n_reps
    res['xicor_perm_stddev'] = numpy.sqrt((1 / (n_reps - 1)) * (
            res['xicor_perm_sum_sq'] - (1 / n_reps) * res['xicor_perm_sum']**2))
    del res['xicor_perm_sum']
    del res['xicor_perm_sum_sq']
    return res
